import csv
from itertools import chain
import numpy as np
from eight_mile.utils import exporter
from collections import OrderedDict


__all__ = []
export = exporter(__all__)


@export
class ConfusionMatrix:
    """Confusion matrix with metrics

    This class accumulates classification output, and tracks it in a confusion matrix.
    Metrics are available that use the confusion matrix
    """

    def __init__(self, labels):
        """Constructor with input labels

        :param labels: Either a dictionary (`k=int,v=str`) or an array of labels
        """
        if isinstance(labels, dict):
            self.labels = []
            for i in range(len(labels)):
                self.labels.append(labels[i])
        else:
            self.labels = labels
        nc = len(self.labels)
        self._cm = np.zeros((nc, nc), dtype=np.int)

    def add(self, truth, guess):
        """Add a single value to the confusion matrix based off `truth` and `guess`

        :param truth: The real `y` value (or ground truth label)
        :param guess: The guess for `y` value (or assertion)
        """

        self._cm[truth, guess] += 1

    def __str__(self):
        values = []
        width = max(8, max(len(x) for x in self.labels) + 1)
        for i, label in enumerate([""] + self.labels):
            values += ["{:>{width}}".format(label, width=width + 1)]
        values += ["\n"]
        for i, label in enumerate(self.labels):
            values += ["{:>{width}}".format(label, width=width + 1)]
            for j in range(len(self.labels)):
                values += ["{:{width}d}".format(self._cm[i, j], width=width + 1)]
            values += ["\n"]
        values += ["\n"]
        return "".join(values)

    def save(self, outfile):
        ordered_fieldnames = OrderedDict([("labels", None)] + [(l, None) for l in self.labels])
        with open(outfile, "w") as f:
            dw = csv.DictWriter(f, delimiter=",", fieldnames=ordered_fieldnames)
            dw.writeheader()
            for index, row in enumerate(self._cm):
                row_dict = {l: row[i] for i, l in enumerate(self.labels)}
                row_dict.update({"labels": self.labels[index]})
                dw.writerow(row_dict)

    def reset(self):
        """Reset the matrix
        """
        self._cm *= 0

    def get_correct(self):
        """Get the diagonals of the confusion matrix

        :return: (``int``) Number of correct classifications
        """
        return self._cm.diagonal().sum()

    def get_total(self):
        """Get total classifications

        :return: (``int``) total classifications
        """
        return self._cm.sum()

    def get_acc(self):
        """Get the accuracy

        :return: (``float``) accuracy
        """
        return float(self.get_correct()) / self.get_total()

    def get_recall(self):
        """Get the recall

        :return: (``float``) recall
        """
        total = np.sum(self._cm, axis=1)
        total = (total == 0) + total
        return np.diag(self._cm) / total.astype(float)

    def get_support(self):
        return np.sum(self._cm, axis=1)

    def get_precision(self):
        """Get the precision
        :return: (``float``) precision
        """

        total = np.sum(self._cm, axis=0)
        total = (total == 0) + total
        return np.diag(self._cm) / total.astype(float)

    def get_mean_precision(self):
        """Get the mean precision across labels

        :return: (``float``) mean precision
        """
        return np.mean(self.get_precision())

    def get_weighted_precision(self):
        return np.sum(self.get_precision() * self.get_support()) / float(self.get_total())

    def get_mean_recall(self):
        """Get the mean recall across labels

        :return: (``float``) mean recall
        """
        return np.mean(self.get_recall())

    def get_weighted_recall(self):
        return np.sum(self.get_recall() * self.get_support()) / float(self.get_total())

    def get_weighted_f(self, beta=1):
        return np.sum(self.get_class_f(beta) * self.get_support()) / float(self.get_total())

    def get_macro_f(self, beta=1):
        """Get the macro F_b, with adjustable beta (defaulting to F1)

        :param beta: (``float``) defaults to 1 (F1)
        :return: (``float``) macro F_b
        """
        if beta < 0:
            raise Exception("Beta must be greater than 0")
        return np.mean(self.get_class_f(beta))

    def get_class_f(self, beta=1):
        p = self.get_precision()
        r = self.get_recall()

        b = beta * beta
        d = b * p + r
        d = (d == 0) + d

        return (b + 1) * p * r / d

    def get_f(self, beta=1):
        """Get 2 class F_b, with adjustable beta (defaulting to F1)

        :param beta: (``float``) defaults to 1 (F1)
        :return: (``float``) 2-class F_b
        """
        p = self.get_precision()[1]
        r = self.get_recall()[1]
        if beta < 0:
            raise Exception("Beta must be greater than 0")
        d = beta * beta * p + r
        if d == 0:
            return 0
        return (beta * beta + 1) * p * r / d

    def get_r_k(self):
        """Calculate the R k correlation coefficient from here
            https://www.sciencedirect.com/science/article/abs/pii/S1476927104000799?via%3Dihub
        See this blog post for an explanation of metric, how this calculation is equivalent and
        how it reduces to MCC for 2 classes.
            www.blester125.com/blog/rk.html
        """
        samples = np.sum(self._cm)
        correct = np.trace(self._cm)
        true = np.sum(self._cm, axis=1, dtype=np.float64)
        guess = np.sum(self._cm, axis=0, dtype=np.float64)

        cov_guess_true = correct * samples - np.dot(guess, true)
        cov_true_true = samples * samples - np.dot(true, true)
        cov_guess_guess = samples * samples - np.dot(guess, guess)

        denom = np.sqrt(cov_guess_guess * cov_true_true)
        denom = denom if denom != 0.0 else 1.0
        return cov_guess_true / denom

    def get_mcc(self):
        """Get the Mathews correlation coefficient.
        R k reduces to the Matthews Correlation Coefficient for two classes.
        People are often more familiar with the name MCC than R k so we provide this alias
        """
        return self.get_r_k()

    def get_all_metrics(self):
        """Make a map of metrics suitable for reporting, keyed by metric name

        :return: (``dict``) Map of metrics keyed by metric names
        """
        metrics = {"acc": self.get_acc()}
        # If 2 class, assume second class is positive AKA 1
        if len(self.labels) == 2:
            metrics["precision"] = self.get_precision()[1]
            metrics["recall"] = self.get_recall()[1]
            metrics["f1"] = self.get_f(1)
            metrics["mcc"] = self.get_mcc()
        else:
            metrics["mean_precision"] = self.get_mean_precision()
            metrics["mean_recall"] = self.get_mean_recall()
            metrics["macro_f1"] = self.get_macro_f(1)
            metrics["weighted_precision"] = self.get_weighted_precision()
            metrics["weighted_recall"] = self.get_weighted_recall()
            metrics["weighted_f1"] = self.get_weighted_f(1)
            metrics["r_k"] = self.get_r_k()
        return metrics

    def add_batch(self, truth, guess):
        """Add a batch of data to the confusion matrix

        :param truth: The truth tensor
        :param guess: The guess tensor
        :return:
        """
        for truth_i, guess_i in zip(truth, guess):
            self.add(truth_i, guess_i)

    @classmethod
    def create(cls, truth, guess):
        """Build a Confusion Matrix from truths and guesses.
        If there are classes missing from the union of golds and preds they will
        be missing from the confusion matrix
        The classes are added in sorted order, if they are ints already this means
        they were probably already encoded and we will hopefully respect the order
        they are in.
        :param truth: List[Union[int. str]] The correct labels
        :param guess: List[Union[int. str]] The predicted labels
        :returns: ConfusionMatrix
        """
        label_index = {i: k for i, k in enumerate(sorted(set(chain(truth, guess))))}
        rev_lut = {v: k for k, v in label_index.items()}
        cm = cls(label_index)
        for g, p in zip(truth, guess):
            cm.add(rev_lut[g], rev_lut[p])
        return cm



from typing import Optional, Tuple
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure


__all__ = ["confidence_histogram"]


def confidence_histogram(
    left_bins: np.array,
    bin_counts: np.array,
    acc: Optional[float] = None,
    avg_conf: Optional[float] = None,
    x_ticks: Optional[np.array] = None,
    y_ticks: Optional[np.array] = None,
    title: Optional[str] = "Confidence Distribution",
    y_label: Optional[str] = "% of Samples",
    x_label: Optional[str] = "Confidence",
    conf_label: Optional[str] = "Avg. Confidence",
    acc_label: Optional[str] = "Accuracy",
    label_y: float = 0.4,
    conf_spacer: float = 0.015,
    acc_spacer: float = -0.07,
    edge: str = "k",
    color: str = "b",
    fig_size: int = 3,
    ax: Optional[Axes] = None,
    fig: Optional[Figure] = None,
) -> Tuple[Figure, Axes]:
    """Plot a reliability graph a la https://arxiv.org/abs/1706.04599

    :param left_bins: The bins that define the groups, they should specify the left
        edge of the bin
    :param accs: The average accuracy for examples in that bin
    :param x_ticks: Where to show the ticks along the x axis. If not provided it
        will display every other tick, pass an empty list to skip adding ticks
    :param y_ticks: Where to show the ticks along the y axis. If not provided it
        uses default matplotlib ticks, padd an empty list to skip adding ticks
    :param title: The title for the axes, pass None to skip adding a title to the axes
    :param y_label: A label for the y axes, pass None to skip adding a label
    :param x_label: A label for the x axes, pass None to skip adding a label
    :param conf_label: A label to put on the line representing the average confidence
    :param acc_label: A label to put on the line representing the accuract
    :param label_y: Where to start the above labels alon the y axis, 0 is the bottom, 1 is the top
    :param conf_spacer: How far off the line to put the confidence label, use positive to move
        right of the line and negative to move to the left
    :param acc_spacer: How far off the line to put the accuracy label, use positive to move right
        of the line and negative to move to the left
    :param edge: The color to use for the edges of the output bins
    :param color: The color to use for the main section of the output bins
    :param fig_size: The figure size, only used if creating the figure/axis from scratch
    :param ax: An Axes object that we can create the graph in, if None it will be created
    :param fig: A figure that holds the axes object

    :returns: `Tuple(fig, ax)` If the figure and ax are passed in they are returned, if they
        were not passed in the created ones are returned. The don't use the figure at all but
        because we want to return both in case we created them.
    """
    # If niether a figure non-axes are passed in we create them.
    if ax is None and fig is None:
        fig, ax = plt.subplots(1, 1, figsize=(fig_size, fig_size))
    # If the axes was not passed in but the figure was we get the current active
    # axes for the figure.
    if ax is None and fig is not None:
        ax = fig.gca()
    # If the axes is provided but the figure we look up the figure on the axes
    if ax is not None and fig is None:
        fig = ax.figure
    widths = np.diff(left_bins, append=1)
    ax.bar(
        left_bins,
        bin_counts / np.sum(bin_counts),
        align='edge',
        width=widths,
        edgecolor=edge,
        color=color
    )
    trans = ax.get_xaxis_transform()
    if avg_conf is not None:
        ax.axvline(avg_conf, ls='--', c='.3')
        if conf_label is not None:
            ax.text(avg_conf + conf_spacer, label_y, conf_label, rotation=90, transform=trans)
    if acc is not None:
        ax.axvline(acc, ls='--', c='.3')
        if acc_label is not None:
            ax.text(acc + acc_spacer, label_y, acc_label, rotation=90, transform=trans)
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.set_aspect("equal", "box")
    ax.set_title(title)
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    if x_ticks is None:
        x_ticks = [y for i, y in enumerate(left_bins) if i % 2 == 0] + [1.0]
    ax.set_xticks(x_ticks)
    if y_ticks is not None:
        ax.set_yticks(y_ticks)
    return fig, ax


def _demo():
    small_bins = np.arange(0, 1, step=0.05)
    counts = np.zeros_like(small_bins)
    prev = 0
    for i in range(len(small_bins)):
        add = np.random.randint(10, 15 * (i + 1))
        counts[i] = prev + add
        prev = counts[i]
    counts[-1] += np.random.randint(50, 1000)
    x_ticks = np.arange(0, 1.1, step=0.2)
    ACC = 0.72
    CONF = 0.84

    f, a = confidence_histogram(small_bins, counts, x_ticks=x_ticks, acc=ACC, avg_conf=CONF)
    plt.show()


if __name__ == "__main__":
    _demo()



import logging
import tensorflow as tf
from eight_mile.utils import get_version, exporter
from eight_mile.optz import create_lr_scheduler, register_lr_scheduler
from eight_mile.optz import (
    LearningRateScheduler,
    ConstantScheduler,
    WarmupLearningRateScheduler,
    WarmupLinearScheduler,
    CyclicLRScheduler,
    PiecewiseDecayScheduler,
    ZarembaDecayScheduler,
    CosineDecayScheduler,
    InverseTimeDecayScheduler,
    ExponentialDecayScheduler,
    CompositeLRScheduler,
)

logger = logging.getLogger("mead.layers")
__all__ = []
export = exporter(__all__)


if not tf.executing_eagerly():

    @register_lr_scheduler("default")
    class ConstantSchedulerTensorFlow1:
        def __init__(self, **kwargs):
            pass

        def __call__(self, lr, global_step):
            return tf.identity(lr, name="lr")

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("warmup_linear")
    class WarmupLinearSchedulerTensorFlow1(WarmupLearningRateScheduler):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def __call__(self, lr, global_step):
            return tf.identity(
                tf.minimum(1.0, tf.cast(global_step / self.warmup_steps, dtype=tf.float32)) * lr, name="lr"
            )

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("clr")
    class CyclicLRSchedulerTensorFlow1:
        def __init__(self, max_lr=1e-2, decay_steps=1000, **kwargs):
            super().__init__()
            self.max_lr = max_lr
            self.decay_steps = decay_steps

        def __call__(self, lr, global_step):
            gs_f = tf.cast(global_step, tf.float32)
            cycle = tf.floor(1.0 + gs_f / (2.0 * self.decay_steps))
            x = tf.abs(gs_f / self.decay_steps - 2.0 * cycle + 1.0)
            clr = lr + (self.max_lr - lr) * tf.maximum(0.0, 1.0 - x)
            return tf.identity(clr, name="lr")

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("sgdr")
    class SGDRSchedulerTensorFlow1:
        def __init__(self, first_decay_steps=1000, **kwargs):
            super().__init__()
            self.first_decay_steps = first_decay_steps

        def __call__(self, lr, global_step):
            return tf.identity(
                tf.train.cosine_decay_restarts(lr, global_step, first_decay_steps=self.first_decay_steps), name="lr"
            )

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("composite")
    class CompositeLRSchedulerTensorFlow1:
        def __init__(self, warm=None, rest=None, **kwargs):
            self.warm = warm
            self.rest = rest

        def __call__(self, lr, global_step):
            warm_tensor = self.warm(lr, global_step)

            def call_warm():
                return warm_tensor

            rest_step = tf.subtract(global_step, tf.compat.v1.constant(self.warm.warmup_steps, dtype=global_step.dtype))
            rest_tensor = self.rest(lr, rest_step)

            def call_rest():
                return rest_tensor

            return tf.identity(tf.cond(global_step < self.warm.warmup_steps, call_warm, call_rest), name="lr")

        def __str__(self):
            return "LRScheduler({}, {})".format(self.warm, self.rest)

    @register_lr_scheduler("piecewise")
    class PiecewiseDecaySchedulerTensorFlow:
        def __init__(self, boundaries=None, values=None, **kwargs):
            super().__init__()
            self.boundaries = boundaries
            self.values = values

        def __call__(self, lr, global_step):
            return tf.identity(tf.compat.v1.train.piecewise_constant(global_step, self.boundaries, self.values), name="lr")

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("invtime")
    class InverseTimeDecaySchedulerTensorFlow:
        def __init__(self, decay_steps=16000, decay_rate=0.05, staircase=False, **kwargs):
            super().__init__()
            self.decay_steps = decay_steps
            self.decay_rate = decay_rate
            self.staircase = staircase

        def __call__(self, lr, global_step):
            return tf.identity(
                tf.train.inverse_time_decay(
                    lr, global_step, self.decay_steps, self.decay_rate, staircase=self.staircase
                ),
                name="lr",
            )

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("exponential")
    class ExponentialDecaySchedulerTensorFlow(object):
        def __init__(self, decay_steps=16000, decay_rate=0.5, staircase=False, **kwargs):
            self.decay_steps = decay_steps
            self.decay_rate = decay_rate
            self.staircase = staircase

        def __call__(self, lr, global_step):
            return tf.identity(
                tf.train.exponential_decay(
                    lr, global_step, self.decay_steps, self.decay_rate, staircase=self.staircase
                ),
                name="lr",
            )

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("zaremba")
    class ZarembaDecaySchedulerTensorFlow1(PiecewiseDecaySchedulerTensorFlow):
        """Utility only, just to simplify the JSON"""

        def __init__(self, boundaries=None, decay_rate=None, **kwargs):
            lr = float(kwargs.get("lr", kwargs.get("eta", 1.0)))
            values = [lr / (float(decay_rate) ** i) for i in range(len(boundaries) + 1)]
            super().__init__(boundaries=boundaries, values=values)

        def __str__(self):
            return type(self).__name__ + "()"


else:

    @register_lr_scheduler("default")
    class ConstantSchedulerTensorFlow2(LearningRateScheduler, tf.keras.optimizers.schedules.LearningRateSchedule):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def __call__(self, global_step):
            return self.lr

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("warmup_linear")
    class WarmupLinearSchedulerTensorFlow2(
        WarmupLearningRateScheduler, tf.keras.optimizers.schedules.LearningRateSchedule
    ):
        def __init__(self, **kwargs):
            lr = float(kwargs.get("lr", kwargs.get("eta", 1.0)))
            kwargs["lr"] = lr
            super().__init__(**kwargs)

        def __call__(self, global_step):
            return tf.minimum(1.0, global_step / float(self.warmup_steps)) * self.lr

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("clr")
    class CyclicLRSchedulerTensorFlow2(LearningRateScheduler, tf.keras.optimizers.schedules.LearningRateSchedule):
        def __init__(self, max_lr=1e-2, decay_steps=1000, **kwargs):
            lr = float(kwargs.get("lr", kwargs.get("eta", 1.0)))
            kwargs["lr"] = lr
            super().__init__(**kwargs)
            self.max_lr = max_lr
            self.decay_steps = decay_steps

        def __call__(self, global_step):
            gs_f = tf.cast(global_step, tf.float32)
            cycle = tf.floor(1.0 + gs_f / (2.0 * self.decay_steps))
            x = tf.abs(gs_f / self.decay_steps - 2.0 * cycle + 1.0)
            clr = self.lr + (self.max_lr - self.lr) * tf.maximum(0.0, 1.0 - x)
            return tf.identity(clr, name="lr")

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("sgdr")
    class SGDRSchedulerTensorFlow2(LearningRateScheduler, tf.keras.optimizers.schedules.LearningRateSchedule):
        def __init__(self, first_decay_steps=1000, **kwargs):
            super().__init__(**kwargs)
            self.first_decay_steps = first_decay_steps

        def __call__(self, global_step):
            return tf.identity(
                tf.compat.v1.train.cosine_decay_restarts(
                    self.lr, global_step, first_decay_steps=self.first_decay_steps
                ),
                name="lr",
            )

        def __str__(self):
            return type(self).__name__ + "()"

    @register_lr_scheduler("zaremba")
    class ZarembaDecaySchedulerTensorFlow2(tf.keras.optimizers.schedules.PiecewiseConstantDecay):
        """Utility only, just to simplify the JSON"""

        def __init__(self, boundaries=None, decay_rate=None, **kwargs):
            lr = float(kwargs.get("lr", kwargs.get("eta", 1.0)))
            values = [lr / (float(decay_rate) ** i) for i in range(len(boundaries) + 1)]
            super().__init__(boundaries, values, kwargs.get("name"))

    @register_lr_scheduler("composite")
    class CompositeLRSchedulerTensorFlow2(CompositeLRScheduler, tf.keras.optimizers.schedules.LearningRateSchedule):
        pass

    @register_lr_scheduler("piecewise")
    class PiecewiseConstantDecayTensorFlow2(tf.keras.optimizers.schedules.PiecewiseConstantDecay):
        def __init__(self, boundaries, values, **kwargs):
            super().__init__(boundaries, values)

    @register_lr_scheduler("invtime")
    class InverseTimeDecayTensorFlow2(tf.keras.optimizers.schedules.InverseTimeDecay):
        def __init__(self, decay_steps=16000, decay_rate=0.05, staircase=False, **kwargs):
            lr = kwargs.get("lr", kwargs.get("eta", 0.01))
            super().__init__(lr, decay_steps, decay_rate, staircase, kwargs.get("name"))

    @register_lr_scheduler("exponential")
    class ExponentialDecayTensorFlow2(tf.keras.optimizers.schedules.ExponentialDecay):
        def __init__(self, decay_steps=16000, decay_rate=0.5, staircase=False, **kwargs):
            lr = kwargs.get("lr", kwargs.get("eta", 0.01))
            super().__init__(lr, decay_steps, decay_rate, staircase, kwargs.get("name"))

    @register_lr_scheduler("cosine")
    class CosineDecayTensorFlow2(CosineDecayScheduler):
        def __init__(self, decay_steps=16000, alpha=0.0, **kwargs):
            kwargs['lr'] = kwargs.get("lr", kwargs.get("eta", 0.01))
            super().__init__(decay_steps, alpha, **kwargs)

        def __call__(self, global_step):
            global_step = tf.math.minimum(global_step, self.decay_steps)
            cosine_decay = 0.5 * (1.0 + tf.cos(3.14159265 * global_step / self.decay_steps))
            decayed = (1 - self.alpha) * cosine_decay + self.alpha
            return self.lr * decayed

@export
class AdamWOptimizer(tf.compat.v1.train.Optimizer):
    """A basic Adam optimizer that includes "correct" L2 weight decay.

    Modified from: https://github.com/google-research/bert/blob/master/optimization.py
    This does the weight decay slightly differently from PyTorch version, putting it before the update
    """

    def __init__(self, learning_rate, weight_decay=0.0, beta_1=0.9, beta_2=0.999, epsilon=1e-6, name="AdamWOptimizer"):
        """Constructs a AdamWOptimizer."""
        super().__init__(False, name)

        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon

    def _get_variable_name(self, param_name):
        import re

        """Get the variable name from the tensor name."""
        m = re.match("^(.*):\\d+$", param_name)
        if m is not None:
            param_name = m.group(1)
        return param_name

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
        assignments = []
        for (grad, param) in grads_and_vars:
            if grad is None or param is None:
                continue

            param_name = self._get_variable_name(param.name)

            m = tf.compat.v1.get_variable(
                name=param_name + "/adam_m",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.compat.v1.zeros_initializer(),
            )
            v = tf.compat.v1.get_variable(
                name=param_name + "/adam_v",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.compat.v1.zeros_initializer(),
            )

            # Standard Adam update.
            next_m = tf.multiply(self.beta_1, m) + tf.multiply(1.0 - self.beta_1, grad)
            next_v = tf.multiply(self.beta_2, v) + tf.multiply(1.0 - self.beta_2, tf.square(grad))

            update = next_m / (tf.sqrt(next_v) + self.epsilon)
            update += self.weight_decay * param
            update_with_lr = self.learning_rate * update
            next_param = param - update_with_lr

            assignments.extend([param.assign(next_param), m.assign(next_m), v.assign(next_v)])
        return tf.group(*assignments, name=name)


OPTIMIZER_SUMMARIES = ["learning_rate", "loss", "gradients", "gradient_norm", "global_gradient_norm"]


def _optimize_loss(
    loss,
    global_step,
    learning_rate,
    optimizer,
    gradient_multipliers=None,
    clip_gradients=None,
    learning_rate_decay_fn=None,
    update_ops=None,
    variables=None,
    name=None,
    summaries=None,
    colocate_gradients_with_ops=False,
    increment_global_step=True,
):
    """Given loss and parameters for optimizer, returns a training op.
    Various ways of passing optimizers include:
    - by string specifying the name of the optimizer. See OPTIMIZER_CLS_NAMES
        for full list. E.g. `optimize_loss(..., optimizer='Adam')`.
    - by function taking learning rate `Tensor` as argument and returning an
        `Optimizer` instance. E.g. `optimize_loss(...,
        optimizer=lambda lr: tf.train.MomentumOptimizer(lr, momentum=0.5))`.
      Alternatively, if `learning_rate` is `None`, the function takes no
      arguments. E.g. `optimize_loss(..., learning_rate=None,
        optimizer=lambda: tf.train.MomentumOptimizer(0.5, momentum=0.5))`.
    - by a subclass of `Optimizer` having a single-argument constructor
        (the argument is the learning rate), such as AdamOptimizer or
        AdagradOptimizer. E.g. `optimize_loss(...,
        optimizer=tf.train.AdagradOptimizer)`.
    - by an instance of a subclass of `Optimizer`.
        E.g., `optimize_loss(..., optimizer=tf.train.AdagradOptimizer(0.5))`.
    Args:
      loss: Scalar `Tensor`.
      global_step: Scalar int `Tensor`, step counter to update on each step
                   unless `increment_global_step` is `False`. If not supplied,
                   it will be fetched from the default graph (see
                   `tf.train.get_global_step` for details). If it has
                   not been created, no step will be incremented with each weight
                   update. `learning_rate_decay_fn` requires `global_step`.
      learning_rate: float or `Tensor`, magnitude of update per each training
                     step. Can be `None`.
      optimizer: string, class or optimizer instance, used as trainer.
                 string should be name of optimizer, like 'SGD',
                   'Adam', 'Adagrad'. Full list in OPTIMIZER_CLS_NAMES constant.
                 class should be sub-class of `tf.Optimizer` that implements
                   `compute_gradients` and `apply_gradients` functions.
                 optimizer instance should be instantiation of `tf.Optimizer`
                   sub-class and have `compute_gradients` and `apply_gradients`
                   functions.
      gradient_noise_scale: float or None, adds 0-mean normal noise scaled by this
                            value.
      gradient_multipliers: dict of variables or variable names to floats.
                            If present, gradients for specified
                            variables will be multiplied by given constant.
      clip_gradients: float, callable or `None`. If float, is provided, a global
        clipping is applied to prevent the norm of the gradient to exceed this
        value. Alternatively, a callable can be provided e.g.: adaptive_clipping.
        This callable takes a `list` of `(gradients, variables)` `tuple`s and
        returns the same thing with the gradients modified.
      learning_rate_decay_fn: function, takes `learning_rate` and `global_step`
                              `Tensor`s, returns `Tensor`.
                              Can be used to implement any learning rate decay
                              functions.
                              For example: `tf.train.exponential_decay`.
                              Ignored if `learning_rate` is not supplied.
      update_ops: list of update `Operation`s to execute at each step. If `None`,
                  uses elements of UPDATE_OPS collection. The order of execution
                  between `update_ops` and `loss` is non-deterministic.
      variables: list of variables to optimize or
                 `None` to use all trainable variables.
      name: The name for this operation is used to scope operations and summaries.
      summaries: List of internal quantities to visualize on tensorboard. If not
                 set, the loss, the learning rate, and the global norm of the
                 gradients will be reported. The complete list of possible values
                 is in OPTIMIZER_SUMMARIES.
      colocate_gradients_with_ops: If True, try colocating gradients with the
                                   corresponding op.
      increment_global_step: Whether to increment `global_step`. If your model
        calls `optimize_loss` multiple times per training step (e.g. to optimize
        different parts of the model), use this arg to avoid incrementing
        `global_step` more times than necessary.
    Returns:
      Training op.
    Raises:
      ValueError: if:
          * `loss` is an invalid type or shape.
          * `global_step` is an invalid type or shape.
          * `learning_rate` is an invalid type or value.
          * `optimizer` has the wrong type.
          * `clip_gradients` is neither float nor callable.
          * `learning_rate` and `learning_rate_decay_fn` are supplied, but no
            `global_step` is available.
          * `gradients` is empty.
    """
    loss = tf.convert_to_tensor(loss)
    if global_step is None:
        global_step = tf.compat.v1.train.get_global_step()

    with tf.compat.v1.variable_scope(name, "OptimizeLoss", [loss, global_step]):
        # Update ops take UPDATE_OPS collection if not provided.
        if update_ops is None:
            update_ops = set(tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.UPDATE_OPS))
        # Make sure update ops are ran before computing loss.
        if update_ops:
            with tf.compat.v1.control_dependencies(list(update_ops)):
                loss = tf.identity(loss)

        # Learning rate variable, with possible decay.
        lr = None
        if learning_rate is not None:
            if isinstance(learning_rate, tf.Tensor) and learning_rate.get_shape().ndims == 0:
                lr = learning_rate
            elif isinstance(learning_rate, float):
                if learning_rate < 0.0:
                    raise ValueError("Invalid learning_rate %s.", learning_rate)
                lr = tf.compat.v1.get_variable(
                    "learning_rate", [], trainable=False, initializer=tf.compat.v1.constant_initializer(learning_rate)
                )
            else:
                raise ValueError(
                    "Learning rate should be 0d Tensor or float. "
                    "Got %s of type %s" % (str(learning_rate), str(type(learning_rate)))
                )
        if summaries is None:
            summaries = ["loss", "learning_rate", "global_gradient_norm"]
        else:
            for summ in summaries:
                if summ not in OPTIMIZER_SUMMARIES:
                    raise ValueError(
                        "Summaries should be one of [%s], you provided %s." % (", ".join(OPTIMIZER_SUMMARIES), summ)
                    )
        if learning_rate is not None and learning_rate_decay_fn is not None:
            if global_step is None:
                raise ValueError("global_step is required for learning_rate_decay_fn.")
            lr = learning_rate_decay_fn(lr, global_step)
            if "learning_rate" in summaries:
                tf.compat.v1.summary.scalar("learning_rate", lr)

        if isinstance(optimizer, type) and issubclass(optimizer, tf.compat.v1.train.Optimizer):
            if lr is None:
                raise ValueError(
                    "Learning rate is None, but should be specified if " "optimizer is class (%s)." % optimizer
                )
            opt = optimizer(learning_rate=lr)
        elif isinstance(optimizer, tf.compat.v1.train.Optimizer):
            opt = optimizer
        elif callable(optimizer):
            if learning_rate is not None:
                opt = optimizer(lr)
            else:
                opt = optimizer()
            if not isinstance(opt, tf.compat.v1.train.Optimizer):
                raise ValueError(
                    "Unrecognized optimizer: function should return " "subclass of Optimizer. Got %s." % str(opt)
                )
        else:
            raise ValueError(
                "Unrecognized optimizer: should be string, "
                "subclass of Optimizer, instance of "
                "subclass of Optimizer or function with one argument. "
                "Got %s." % str(optimizer)
            )

        # All trainable variables, if specific variables are not specified.
        if variables is None:
            variables = tf.compat.v1.trainable_variables()

        # Compute gradients.
        gradients = opt.compute_gradients(loss, variables, colocate_gradients_with_ops=colocate_gradients_with_ops)

        # Multiply some gradients.
        if gradient_multipliers is not None:
            gradients = _multiply_gradients(gradients, gradient_multipliers)
            if not gradients:
                raise ValueError(
                    "Empty list of (gradient, var) pairs encountered. This is most "
                    "likely to be caused by an improper value of gradient_multipliers."
                )

        if "global_gradient_norm" in summaries or "gradient_norm" in summaries:
            tf.compat.v1.summary.scalar("global_norm/gradient_norm", tf.linalg.global_norm(list(zip(*gradients))[0]))

        # Optionally clip gradients by global norm.
        if isinstance(clip_gradients, float):
            gradients = _clip_gradients_by_norm(gradients, clip_gradients)
        elif callable(clip_gradients):
            gradients = clip_gradients(gradients)
        elif clip_gradients is not None:
            raise ValueError("Unknown type %s for clip_gradients" % type(clip_gradients))

        # Add scalar summary for loss.
        if "loss" in summaries:
            tf.compat.v1.summary.scalar("loss", loss)

        # Add histograms for variables, gradients and gradient norms.
        for gradient, variable in gradients:
            if isinstance(gradient, tf.IndexedSlices):
                grad_values = gradient.values
            else:
                grad_values = gradient

            if grad_values is not None:
                var_name = variable.name.replace(":", "_")
                if "gradients" in summaries:
                    tf.compat.v1.summary.histogram("gradients/%s" % var_name, grad_values)
                if "gradient_norm" in summaries:
                    tf.compat.v1.summary.scalar("gradient_norm/%s" % var_name, tf.linalg.global_norm([grad_values]))

        if clip_gradients is not None and ("global_gradient_norm" in summaries or "gradient_norm" in summaries):
            tf.compat.v1.summary.scalar(
                "global_norm/clipped_gradient_norm", tf.linalg.global_norm(list(zip(*gradients))[0])
            )

        # Create gradient updates.
        grad_updates = opt.apply_gradients(
            gradients, global_step=global_step if increment_global_step else None, name="train"
        )

        # Ensure the train_tensor computes grad_updates.
        # train_tensor = tf.compat.v1.with_dependencies([grad_updates], loss)
        with tf.control_dependencies([grad_updates]):
            train_tensor = tf.identity(loss)

        return train_tensor


def _clip_gradients_by_norm(grads_and_vars, clip_gradients):
    """Clips gradients by global norm."""
    gradients, variables = zip(*grads_and_vars)
    clipped_gradients, _ = tf.clip_by_global_norm(gradients, clip_gradients)
    return list(zip(clipped_gradients, variables))


def _multiply_gradients(grads_and_vars, gradient_multipliers):
    """Multiply specified gradients."""
    multiplied_grads_and_vars = []
    for grad, var in grads_and_vars:
        if grad is not None and (var in gradient_multipliers or var.name in gradient_multipliers):
            key = var if var in gradient_multipliers else var.name
            multiplier = gradient_multipliers[key]
            if isinstance(grad, tf.IndexedSlices):
                grad_values = grad.values * multiplier
                grad = tf.IndexedSlices(grad_values, grad.indices, grad.dense_shape)
            else:
                grad *= tf.cast(multiplier, grad.dtype)
        multiplied_grads_and_vars.append((grad, var))
    return multiplied_grads_and_vars


@export
def optimizer(loss_fn, **kwargs):

    global_step = tf.compat.v1.train.get_or_create_global_step()
    clip = kwargs.get("clip", None)
    optim = kwargs.get("optim", "sgd")
    eta = kwargs.get("lr", kwargs.get("eta", 0.01))
    lr_scheduler = create_lr_scheduler(**kwargs)
    colocate_gradients_with_ops = bool(kwargs.get("colocate_gradients_with_ops", False))
    sgd_mom = float(kwargs.get("mom", 0.9))
    if optim == "adadelta":
        rho = float(kwargs.get("rho", 0.95))
        eps = float(kwargs.get("epsilon", 1e-6))
        logger.info("adadelta(eta=%f, rho=%f, epsilon=%f)", eta, rho, eps)
        optz = lambda lr: tf.compat.v1.train.AdadeltaOptimizer(lr, rho, eps)
    elif optim == "adam":
        beta1 = float(kwargs.get("beta1", 0.9))
        beta2 = float(kwargs.get("beta2", 0.999))
        eps = float(kwargs.get("epsilon", 1e-8))
        logger.info("adam(eta=%f beta1=%f, beta2=%f, eps=%f)", eta, beta1, beta2, eps)
        optz = lambda lr: tf.compat.v1.train.AdamOptimizer(lr, beta1, beta2, eps)
    elif optim == "adamw":
        wd = float(kwargs.get("weight_decay", 0))
        beta1 = float(kwargs.get("beta1", 0.9))
        beta2 = float(kwargs.get("beta2", 0.999))
        eps = float(kwargs.get("epsilon", 1e-8))
        logger.info("adamw(eta=%f beta1=%f, beta2=%f, eps=%f, wd=%f)", eta, beta1, beta2, eps, wd)
        optz = lambda lr: AdamWOptimizer(lr, wd, beta1, beta2, eps)
    elif optim == "rmsprop":
        # Get mom again with difference default
        mom = float(kwargs.get("mom", 0.0))
        logger.info("rmsprop(eta=%f, mom=%f)", eta, mom)
        optz = lambda lr: tf.compat.v1.train.RMSPropOptimizer(lr, momentum=mom)
    elif sgd_mom > 0:
        logger.info("sgd-mom(eta=%f, mom=%f)", eta, sgd_mom)
        optz = lambda lr: tf.compat.v1.train.MomentumOptimizer(lr, sgd_mom)
    else:
        logger.info("sgd(eta=%f)", eta)
        optz = lambda lr: tf.compat.v1.train.GradientDescentOptimizer(lr)

    logger.info("clip gradients at %s", clip)
    return (
        global_step,
        _optimize_loss(
            loss_fn,
            global_step,
            eta,
            optz,
            colocate_gradients_with_ops=colocate_gradients_with_ops,
            clip_gradients=clip,
            learning_rate_decay_fn=lr_scheduler,
            increment_global_step=True,
            variables=kwargs.get('variables')
        ),
    )


@export
class EagerOptimizer(object):
    def __init__(self, loss, optimizer=None, **kwargs):
        self.loss = loss
        self.global_step = tf.Variable(int(kwargs.get('global_step', 0)))
        if "lr_function" in kwargs:
            lr_function = kwargs["lr_function"]
        else:
            if "lr_scheduler_type" not in kwargs:
                kwargs["lr_scheduler_type"] = "default"
            lr_function = create_lr_scheduler(**kwargs)
        # decay_fn = None
        # Right now this option is pointless since sparse updates dont work on GPU.  We just turn it off
        sgd_mom = float(kwargs.get("mom", 0.9))
        self.clip = kwargs.get("clip", 100)

        if optimizer:
            self.optimizer = optimizer
        else:
            optim = kwargs.get("optim", "sgd")
            lr = kwargs.get("lr", kwargs.get("eta", 0.01))

            if optim == "adadelta":
                rho = float(kwargs.get("rho", 0.95))
                eps = float(kwargs.get("epsilon", 1e-6))
                logger.info("adadelta(eta=%f, rho=%f, epsilon=%f)", lr, rho, eps)
                self.optimizer = tf.keras.optimizers.Adadelta(lr, rho, eps)
            elif optim == "adam":
                beta1 = float(kwargs.get("beta1", 0.9))
                beta2 = float(kwargs.get("beta2", 0.999))
                eps = float(kwargs.get("epsilon", 1e-8))
                logger.info("adam(eta=%f beta1=%f, beta2=%f, eps=%f)", lr, beta1, beta2, eps)
                self.optimizer = tf.keras.optimizers.Adam(lr_function, beta1, beta2, eps)

            elif optim == "adamw":
                import tensorflow_addons as tfa

                wd = float(kwargs.get("weight_decay", 0))
                beta1 = float(kwargs.get("beta1", 0.9))
                beta2 = float(kwargs.get("beta2", 0.999))
                eps = float(kwargs.get("epsilon", 1e-8))
                logger.info("adamw(eta=%f beta1=%f, beta2=%f, eps=%f, wd=%f)", lr, beta1, beta2, eps, wd)
                self.optimizer = tfa.optimizers.AdamW(
                    weight_decay=wd, learning_rate=lr_function, beta_1=beta1, beta_2=beta2, epsilon=eps
                )
            elif optim == "rmsprop":
                # Get mom again with difference default
                mom = float(kwargs.get("mom", 0.0))
                logger.info("rmsprop(eta=%f, mom=%f)", lr, mom)
                self.optimizer = tf.keras.optimizers.RMSprop(lr_function, momentum=mom)
            elif sgd_mom > 0:
                logger.info("sgd-mom(eta=%f, mom=%f)", lr, sgd_mom)
                self.optimizer = tf.keras.optimizers.SGD(lr_function, sgd_mom)
            else:
                logger.info("sgd(eta=%f)", lr)
                self.optimizer = tf.keras.optimizers.SGD(lr_function)

        logger.info("clip gradients at %s", self.clip)

    def update(self, model, x, y, num_replicas=1):
        with tf.GradientTape() as tape:
            loss_value = self.loss(model, x, y) / num_replicas
        grads = tape.gradient(loss_value, model.trainable_variables)
        grads, _ = tf.clip_by_global_norm(grads, self.clip)
        self.optimizer.apply_gradients(zip(grads, model.trainable_variables))
        self.global_step.assign_add(1)
        return loss_value

    def update_with_hidden(self, model, h, x, y):
        with tf.GradientTape() as tape:
            loss_value, h = self.loss(model, h, x, y)

        grads = tape.gradient(loss_value, model.trainable_variables)
        grads, _ = tf.clip_by_global_norm(grads, self.clip)
        self.optimizer.apply_gradients(zip(grads, model.trainable_variables))
        self.global_step.assign_add(1)
        return loss_value, h



import math
from collections import OrderedDict
import numpy as np
import torch
import torch.nn as nn
from eight_mile.utils import Offsets, is_sequence, calc_nfeats
from eight_mile.pytorch.layers import *


class PyTorchEmbeddings(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

    def get_vsz(self):
        pass

    def get_dsz(self):
        pass

    @property
    def output_dim(self):
        return self.get_dsz()

    def encode(self, x):
        return self(x)


class LookupTableEmbeddings(PyTorchEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.vsz = kwargs.get("vsz")
        self.dsz = kwargs.get("dsz")
        self.finetune = kwargs.get("finetune", True)
        self.dropin = kwargs.get("dropin", 0.0)

        weights = kwargs.get("weights")
        if weights is None:
            self.embeddings = nn.Embedding(self.vsz, self.dsz, padding_idx=Offsets.PAD)
        else:
            self.embeddings = pytorch_embedding(weights, self.finetune)
            # This makes sure that if you init with a weight and not vsz it will still be available
            self.vsz, self.dsz = weights.shape

    def get_vsz(self):
        return self.vsz

    def get_dsz(self):
        return self.dsz

    def forward(self, x):
        if not self.dropin:
            return self.embeddings(x)

        mask = self.embeddings.weight.data.new().resize_((self.embeddings.weight.size(0),
                                                          1)).bernoulli_(1 - self.dropin).expand_as(self.embeddings.weight) / (1 - self.dropin)
        masked_embed_weight = mask * self.embeddings.weight
        output = torch.nn.functional.embedding(x, masked_embed_weight,
                                               self.embeddings.padding_idx, self.embeddings.max_norm, self.embeddings.norm_type,
                                               self.embeddings.scale_grad_by_freq, self.embeddings.sparse)
        return output

    def extra_repr(self):
        return f"finetune=False" if not self.finetune else ""


class CharConvEmbeddings(PyTorchEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.nfeat_factor = kwargs.get("nfeat_factor")
        self.cfiltsz = kwargs.get("cfiltsz", kwargs.get("filtsz", [3]))
        self.max_feat = kwargs.get("max_feat", 30)
        self.gating = kwargs.get("gating", "skip")
        self.num_gates = kwargs.get("num_gates", 1)
        self.activation = kwargs.get("activation", "tanh")
        self.wsz = kwargs.get("wsz", 30)
        self.projsz = kwargs.get("projsz", 0)
        self.pdrop = kwargs.get("pdrop", 0.5)
        self.filtsz, self.nfeats = calc_nfeats(self.cfiltsz, self.nfeat_factor, self.max_feat, self.wsz)
        self.conv_outsz = int(np.sum(self.nfeats))
        self.outsz = self.conv_outsz
        if self.projsz > 0:
            self.outsz = self.projsz
        self.proj = pytorch_linear(self.conv_outsz, self.outsz)

        self.embeddings = LookupTableEmbeddings(**kwargs)
        self.char_comp = WithDropout(
            ParallelConv(self.embeddings.output_dim, self.nfeats, self.filtsz, self.activation), self.pdrop
        )

        GatingConnection = SkipConnection if self.gating == "skip" else Highway
        self.gating_seq = nn.Sequential(
            OrderedDict(
                [("gate-{}".format(i), GatingConnection(self.char_comp.output_dim)) for i in range(self.num_gates)]
            )
        )

    def get_dsz(self):
        return self.outsz

    def get_vsz(self):
        return self.vsz

    def forward(self, xch):

        # For starters we need to perform embeddings for each character
        # (TxB) x W -> (TxB) x W x D
        _0, _1, W = xch.shape
        char_vecs = self.embeddings(xch.view(-1, W))
        # (TxB) x D x W
        # char_vecs = char_embeds.transpose(1, 2).contiguous()

        #        pytorch_activation(self.activation_type)
        mots = self.char_comp(char_vecs)
        gated = self.gating_seq(mots)
        if self.projsz:
            gated = self.proj(gated)
        return gated.view(_0, _1, self.get_dsz())


class CharLSTMEmbeddings(PyTorchEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.embed = LookupTableEmbeddings(**kwargs)
        self.lstmsz = kwargs.get("lstmsz", 50)
        layers = kwargs.get("layers", 1)
        pdrop = kwargs.get("pdrop", 0.5)
        unif = kwargs.get("unif", 0)
        weight_init = kwargs.get("weight_init", "uniform")
        self.char_comp = BiLSTMEncoderHidden(
            self.embed.output_dim, self.lstmsz, layers, pdrop, unif=unif, initializer=weight_init
        )

    def forward(self, xch):
        B, T, W = xch.shape
        flat_chars = xch.view(-1, W)
        char_embeds = self.embed(flat_chars)

        # Calculate the lengths of each word
        lengths = torch.sum(flat_chars != Offsets.PAD, dim=1)

        # Sort the input to appease the cuDNN gods
        sorted_word_lengths, perm_idx = lengths.sort(0, descending=True)
        sorted_feats = char_embeds[perm_idx].transpose(0, 1).contiguous()

        # cuDNN throws an error if there is an input with a length of 0, this happens when the "word"
        # is actually a "<PAD>" so there are no characters to run the LSTM over. Here we just say
        # that the lengths is 1. This will make cudnn happy and we will just get junk in that spot
        patched_lengths = sorted_word_lengths.masked_fill(sorted_word_lengths == 0, 1)

        # Run the LSTM
        hidden = self.char_comp((sorted_feats, patched_lengths))

        # Create a mask that is true when the sorted length is 0 (where the word was a pad) so that
        # we can mask out the junk that the lstm created because we needed a length of 1
        hidden = hidden.masked_fill((sorted_word_lengths == 0).unsqueeze(-1), 0)

        # Undo the sort so that the representations of the words are in the correct part of the sentence.
        results = unsort_batch(hidden, perm_idx)

        return results.reshape((B, T, -1))

    def get_dsz(self):
        return self.lstmsz

    def get_vsz(self):
        return self.embed.get_vsz()


class CharTransformerEmbeddings(PyTorchEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.embed = LookupTableEmbeddings(**kwargs)
        self.d_model = kwargs.get("wsz", 30)
        self.num_heads = kwargs.get("num_heads", 3)
        self.rpr_k = kwargs.get("rpr_k", 10)
        layers = kwargs.get("layers", 1)
        pdrop = kwargs.get("pdrop", 0.5)
        self.char_comp = TransformerEncoderStackWithLengths(
            self.num_heads, self.d_model, pdrop, False, layers, rpr_k=self.rpr_k, input_sz=self.embed.output_dim
        )

    def forward(self, xch):
        B, T, W = xch.shape
        flat_chars = xch.view(-1, W)
        char_embeds = self.embed(flat_chars)

        # Calculate the lengths of each word
        lengths = torch.sum(flat_chars != Offsets.PAD, dim=1)
        results = self.char_comp((char_embeds, lengths))
        # B,T,H output, how to pool this
        pooled = torch.max(results, -2, keepdims=False)[0]
        return pooled.reshape((B, T, -1))

    def get_dsz(self):
        return self.d_model

    def get_vsz(self):
        return self.embed.get_vsz()


class PositionalMixin(nn.Module):
    """A Mixin that provides functionality to generate positional embeddings to be added to the normal embeddings.

    Note, mixins need to be before the base case when used, i.e.
        `Embedding(Mixin, BaseEmbed)` NOT `Embedding(BaseEmbed, Mixin)`
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def positional(self, length):
        pass

    def extra_repr(self):
        return f"mxlen={self.mxlen}"


class SinusoidalPositionalMixin(PositionalMixin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # This could get us in trouble, if in doubt, pick something big
        self.mxlen = kwargs.get("mxlen", 1000)
        max_timescale = kwargs.get("max_timescale", 1.0e4)

        word_dsz = self.get_dsz()

        log_timescale_increment = math.log(max_timescale) / word_dsz
        inv_timescales = torch.exp(torch.arange(0, word_dsz, 2).float() * -log_timescale_increment)

        pe = torch.zeros(self.mxlen, word_dsz)
        position = torch.arange(0, self.mxlen).float().unsqueeze(1)
        pe[:, 0::2] = torch.sin(position * inv_timescales)
        pe[:, 1::2] = torch.cos(position * inv_timescales)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def positional(self, length):
        return self.pe[:, :length]


class LearnedPositionalMixin(PositionalMixin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mxlen = int(kwargs.get("mxlen", 512))
        self.pos_embeddings = nn.Embedding(self.mxlen, self.get_dsz())

    def positional(self, length):
        return self.pos_embeddings(
            torch.arange(length, dtype=torch.long, device=self.pos_embeddings.weight.device)
        ).unsqueeze(0)


class BERTLookupTableEmbeddings(LookupTableEmbeddings):
    """
    BERT style embeddings with a 0 token type

    TODO: Get rid of this, we dont need it anymore
    If you want to use BERT with token types, make a `LearnedPositionalLookupTableEmbeddings` feature
    and a `LookupTableEmbeddings` feature (for the token type)
    and put them in an `EmbeddingsStack` with an embeddings_reduction='sum-layer-norm' on the model

    Otherwise, if you do not plan on setting the token type, use the `LearnedPositionalLookupTableEmbeddingsWithBias`,
    which will add the BERT token_type=0 weights into the pos + word_embed and is more efficient
    than this class, since it doesnt do any memory allocation on the fly
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get('dropout', 0.1))
        self.mxlen = int(kwargs.get('mxlen', 512))
        self.tok_type_vsz = kwargs['tok_type_vsz']
        self.pos_embeddings = nn.Embedding(self.mxlen, self.get_dsz())
        self.tok_embeddings = nn.Embedding(self.tok_type_vsz, self.get_dsz())
        self.ln = nn.LayerNorm(self.get_dsz(), eps=1e-12)

    def forward(self, x):
        zeros = torch.zeros_like(x)
        x = super().forward(x)
        x = x + self.positional(x.size(1)) + self.tok_embeddings(zeros)
        x = self.ln(x)
        return self.dropout(x)

    def positional(self, length):
        return self.pos_embeddings(
            torch.arange(length, dtype=torch.long, device=self.pos_embeddings.weight.device)
        ).unsqueeze(0)


class LearnedPositionalLookupTableEmbeddingsWithBias(LookupTableEmbeddings):
    """Learned positional lookup table embeddings wih a bias and layer norm

    This is just a typical learned positional embedding but with a learnable
    bias and a layer norm.  This is equivalent to BERT embeddings when the
    token_type is not set.

    If you are using BERT but you have no interest in using token type embeddings
    (IOW if you are setting all the values of that feature zero anyhow), using this
    object is faster and simpler than having a separate vectorizer for token type.

    If you have a need for token type embeddings, you will want to create 2 sets of embeddings,
    one that acts on the tokens, of type `LearnedPositionalLookupTableEmbeddings` and one of the type
    `LookupTableEmbeddings` for the token type feature

    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get('dropout', 0.1))
        self.mxlen = int(kwargs.get('mxlen', 512))
        self.pos_embeddings = nn.Embedding(self.mxlen, self.get_dsz())
        self.bias = nn.Parameter(torch.zeros(self.get_dsz()))

    def forward(self, x):
        x = super().forward(x)
        x = x + self.positional(x.size(1)) + self.bias
        return x

    def positional(self, length):
        return self.pos_embeddings(
            torch.arange(length, dtype=torch.long, device=self.pos_embeddings.weight.device)
        ).unsqueeze(0)


class PositionalLookupTableEmbeddings(SinusoidalPositionalMixin, LookupTableEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.1))
        self.scale = math.sqrt(self.get_dsz())

    def forward(self, x):
        """Add a positional encoding to the embedding, followed by dropout

        :param x: The temporal signal in, to which the positional embeddings are applied
        :return: Embedded output
        """
        x = super().forward(x) * self.scale
        x = x + self.positional(x.size(1))
        return self.dropout(x)


class LearnedPositionalLookupTableEmbeddings(LearnedPositionalMixin, LookupTableEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.1))

    def forward(self, x):
        T = x.size(1)
        x = super().forward(x)
        pos = self.positional(T)
        return self.dropout(x + pos)


class PositionalCharConvEmbeddings(SinusoidalPositionalMixin, CharConvEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.1))
        self.scale = math.sqrt(self.get_dsz())

    def forward(self, xch):
        """Add a positional encoding to the embedding, followed by dropout

        :param xch: The temporal signal in, to which the positional embeddings are applied
        :return: Embedded output
        """
        xch = super().forward(xch) * self.scale
        xch = xch + self.positional(xch.size(1))
        return self.dropout(xch)


class LearnedPositionalCharConvEmbeddings(LearnedPositionalMixin, CharConvEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.1))

    def forward(self, xch):
        """Add a positional encoding to the embedding, followed by dropout

        :param xch: The temporal signal in, to which the positional embeddings are applied
        :return: Embedded output
        """
        xch = super().forward(xch)
        xch = xch + self.positional(xch.size(1))
        return self.dropout(xch)


class PositionalCharLSTMEmbeddings(SinusoidalPositionalMixin, CharLSTMEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.1))
        self.scale = math.sqrt(self.get_dsz())

    def forward(self, xch):
        xch = super().forward(xch) * self.scale
        xch = xch + self.positional(xch.size(1))
        return self.dropout(xch)


class LearnedPositionalCharLSTMEmbeddings(LearnedPositionalMixin, CharLSTMEmbeddings):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.1))

    def forward(self, xch):
        xch = super().forward(xch)
        xch = xch + self.positional(xch.size(1))
        return self.dropout(xch)



import numpy as np


__all__ = ["expected_calibration_error", "maximum_calibration_error"]


def expected_calibration_error(
    bin_mean_acc: np.ndarray,
    bin_mean_conf: np.ndarray,
    bin_counts: np.ndarray,
) -> float:
    """The difference in expectation between the confidence and the accuracy.

    This is an approximation of eq (2) as described in eq (5)
        from https://arxiv.org/abs/1706.04599

    This is the absolute difference between the confidence and accuracy of each
    bin weighted by the number of samples in each bin.

    :param bin_mean_acc: `[B]` an array with the accuracy of each bin as elements
    :param bin_meanc_conf: `[B]` an array with the mean confidence of each bin as elements
    :param bin_counts: `[B]` an array with the number of samples in each bin

    :returns: The ECE between the accuracy and confidence distributions via binning.
    """
    abs_differences = np.abs(bin_mean_acc - bin_mean_conf)
    coeff = bin_counts / np.sum(bin_counts)
    return np.sum(coeff * abs_differences)


def maximum_calibration_error(
    bin_mean_acc: np.ndarray,
    bin_mean_conf: np.ndarray,
    bin_counts: np.ndarray,
) -> float:
    """The worst case deviation between confidence and accuracy.

    This is an approximation of eq (4) as described in eq (5)
        from https://arxiv.org/abs/1706.04599

    This is the maximum absolute difference between the confidence and accuracy
    of each bin. The bin_counts are used to filter out bins that have no samples
    in them. This seems like it is very dependent on the number of bins you have
    and should therefore be kept constant when comparing models.

    :param bin_mean_acc:
    :param bin_mean_conf:
    :param bin_counts:

    :returns: The MCE between the accuracy and confidence distributions via binning.
    """
    abs_differences = np.abs(bin_mean_acc - bin_mean_conf)
    filtered = abs_differences[bin_counts != 0]
    return np.max(filtered)



import re
import io
import os
import sys
import json
import logging
import inspect
import importlib
import collections
from itertools import chain
from binascii import hexlify
from functools import partial
from collections import Counter
from typing import List, Tuple, Union, Optional, Dict, Any, Set, Pattern
from functools import partial, update_wrapper, wraps
import numpy as np
from six.moves.urllib.request import urlretrieve


logger = logging.getLogger("mead.layers")

__all__ = ["exporter", "optional_params", "parameterize"]
CONLL_EXTS = {".bio", ".iobes", ".iob", ".conll"}


def optional_params(func):
    """Allow a decorator to be called without parentheses if no kwargs are given.

    parameterize is a decorator, function is also a decorator.
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        """If a decorator is called with only the wrapping function just execute the real decorator.
           Otherwise return a lambda that has the args and kwargs partially applied and read to take a function as an argument.

        *args, **kwargs are the arguments that the decorator we are parameterizing is called with.

        the first argument of *args is the actual function that will be wrapped
        """
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return func(args[0])
        return lambda x: func(x, *args, **kwargs)

    return wrapped


def parameterize(func):
    """Allow as decorator to be called with arguments, returns a new decorator that should be called with the function to be wrapped."""

    @wraps(func)
    def decorator(*args, **kwargs):
        return lambda x: func(x, *args, **kwargs)

    return decorator


@parameterize
def exporter(obj, all_list: List[str] = None):
    """Add a function or class to the __all__.

    When exporting something with out using as a decorator do it like so:
        `func = exporter(func)`
    """
    all_list.append(obj.__name__)
    return obj


export = exporter(__all__)


@export
def register(cls, registry, name=None, error=""):
    if name is None:
        name = cls.__name__
    if name in registry:
        raise Exception(
            "Error: attempt to re-define previously registered {} {} (old: {}, new: {})".format(
                error, name, registry[name], cls
            )
        )
    if hasattr(cls, "create"):
        registry[name] = cls.create
    else:
        registry[name] = cls
    return cls


@export
class Offsets:
    """Support pre 3.4"""

    PAD, GO, EOS, UNK, OFFSET = range(0, 5)
    VALUES = ["<PAD>", "<GO>", "<EOS>", "<UNK>"]


@export
def get_logging_level(level: str) -> int:
    """Get the logging level as a logging module constant.

    :param level: `str` The log level to get.

    :returns: The log level, defaults to `INFO`
    """
    return getattr(logging, level.upper(), logging.INFO)


@export
def sequence_mask(lengths, max_len: int = -1):
    if max_len < 0:
        max_len = np.max(lengths)
    row = np.arange(0, max_len).reshape(1, -1)
    col = np.reshape(lengths, (-1, 1))
    return (row < col).astype(np.uint8)


@export
def calc_nfeats(
    filtsz: Union[List[Tuple[int, int]], List[int]],
    nfeat_factor: Optional[int] = None,
    max_feat: Optional[int] = None,
    nfeats: Optional[int] = None,
) -> Tuple[List[int], List[int]]:
    """Calculate the output sizes to use for multiple parallel convolutions.

    If filtsz is a List of Lists of ints then we assume that each element represents
        a filter size, feature size pair. This is the format used by ELMo
    If filtsz is a List of ints we assume each element represents a filter size
    If nfeat_factor and max_feat are set we calculate the nfeat size based on the
        nfeat_factor and the filter size capped by max_feat. This is the method used
        in Kim et. al. 2015 (https://arxiv.org/abs/1508.06615)
    Otherwise nfeats must be set and we assume this is output size to use for all of
        the parallel convs and return the feature size expanded to list the same length
        as filtsz

    :param filtsz: The filter sizes to use in parallel
    :param nfeat_factor: How to scale the feat size as you grow the filters
    :param max_feat: The cap on the feature size
    :param nfeats: A fall back constant feature size
    :returns: Associated arrays where the first one is the filter sizes and the second
        one has the corresponding number of feats as the output
    """
    # If this is a list, then its a tuple of (filtsz, nfeats)
    if is_sequence(filtsz[0]):
        filtsz, nfeats = zip(*filtsz)
    # If we get a nfeat factor, we multiply that by each filter, and thresh at max_feat
    elif nfeat_factor is not None:
        assert max_feat is not None, "If using `nfeat_factor`, `max_feat` must not be None"
        nfeats = [min(nfeat_factor * fsz, max_feat) for fsz in filtsz]
    # Otherwise its just a scalar
    else:
        assert nfeats is not None, "When providing only `filtsz` and not `nfeat_factor` `nfeats` must be specified"
        assert isinstance(
            nfeats, int
        ), "If you want to use custom nfeat sizes do `filtsz = zip(filtsz, nfeats)` then call this function"
        nfeats = [nfeats] * len(filtsz)
    return filtsz, nfeats


@export
def transition_mask(vocab: Dict[str, int], span_type: str, s_idx: int, e_idx: int, pad_idx: Optional[int] = None):
    """Create a CRF mask.

    Returns a mask with invalid moves as 0 and valid as 1.

    :param vocab: dict, Label vocabulary mapping name to index.
    :param span_type: str, The sequence labeling formalism {IOB, IOB2, BIO, or IOBES}
    :param s_idx: int, What is the index of the GO symbol?
    :param e_idx: int, What is the index of the EOS symbol?
    :param pad_idx: int, What is the index of the PAD symbol?

    Note:
        In this mask the PAD symbol is between the last symbol and EOS, PADS can
        only move to pad and the EOS. Any symbol that can move to an EOS can also
        move to a pad.
    """
    rev_lut = {v: k for k, v in vocab.items()}
    start = rev_lut[s_idx]
    end = rev_lut[e_idx]
    pad = None if pad_idx is None else rev_lut[pad_idx]
    if span_type.upper() == "IOB":
        mask = iob_mask(vocab, start, end, pad)
    if span_type.upper() in ("IOB2", "BIO"):
        mask = iob2_mask(vocab, start, end, pad)
    if span_type.upper() == "IOBES":
        mask = iobes_mask(vocab, start, end, pad)
    return mask


@export
def iob_mask(vocab, start, end, pad=None):
    small = 0
    mask = np.ones((len(vocab), len(vocab)), dtype=np.float32)
    for from_ in vocab:
        for to in vocab:
            # Can't move to start
            if to is start:
                mask[vocab[to], vocab[from_]] = small
            # Can't move from end
            if from_ is end:
                mask[vocab[to], vocab[from_]] = small
            # Can only move from pad to pad or end
            if from_ is pad:
                if not (to is pad or to is end):
                    mask[vocab[to], vocab[from_]] = small
            elif from_ is start:
                # Can't move from start to a B
                if to.startswith("B-"):
                    mask[vocab[to], vocab[from_]] = small
            else:
                if from_.startswith("B-"):
                    # Can't move from a B to a B of another type
                    if to.startswith("B-"):
                        from_type = from_.split("-", maxsplit=1)[1]
                        to_type = to.split("-", maxsplit=1)[1]
                        if from_type != to_type:
                            mask[vocab[to], vocab[from_]] = small
                elif from_.startswith("I-"):
                    # Can't move from an I to a B of another type
                    if to.startswith("B-"):
                        from_type = from_.split("-", maxsplit=1)[1]
                        to_type = to.split("-", maxsplit=1)[1]
                        if from_type != to_type:
                            mask[vocab[to], vocab[from_]] = small
                elif from_.startswith("O"):
                    # Can't move from an O to a B
                    if to.startswith("B-"):
                        mask[vocab[to], vocab[from_]] = small
    return mask


@export
def iob2_mask(vocab, start, end, pad=None):
    small = 0
    mask = np.ones((len(vocab), len(vocab)), dtype=np.float32)
    for from_ in vocab:
        for to in vocab:
            # Can't move to start
            if to is start:
                mask[vocab[to], vocab[from_]] = small
            # Can't move from end
            if from_ is end:
                mask[vocab[to], vocab[from_]] = small
            # Can only move from pad to pad or end
            if from_ is pad:
                if not (to is pad or to is end):
                    mask[vocab[to], vocab[from_]] = small
            elif from_ is start:
                # Can't move from start to a I
                if to.startswith("I-"):
                    mask[vocab[to], vocab[from_]] = small
            else:
                if from_.startswith("B-"):
                    # Can't move from a B to an I of a different type
                    if to.startswith("I-"):
                        from_type = from_.split("-", maxsplit=1)[1]
                        to_type = to.split("-", maxsplit=1)[1]
                        if from_type != to_type:
                            mask[vocab[to], vocab[from_]] = small
                elif from_.startswith("I-"):
                    # Can't move from an I to an I of another type
                    if to.startswith("I-"):
                        from_type = from_.split("-", maxsplit=1)[1]
                        to_type = to.split("-", maxsplit=1)[1]
                        if from_type != to_type:
                            mask[vocab[to], vocab[from_]] = small
                elif from_.startswith("O"):
                    # Can't move from an O to an I
                    if to.startswith("I-"):
                        mask[vocab[to], vocab[from_]] = small
    return mask


@export
def iobes_mask(vocab, start, end, pad=None):
    small = 0
    mask = np.ones((len(vocab), len(vocab)), dtype=np.float32)
    for from_ in vocab:
        for to in vocab:
            # Can't move to start
            if to is start:
                mask[vocab[to], vocab[from_]] = small
            # Can't move from end
            if from_ is end:
                mask[vocab[to], vocab[from_]] = small
            # Can only move from pad to pad or to end
            if from_ is pad:
                if not (to is pad or to is end):
                    mask[vocab[to], vocab[from_]] = small
            elif from_ is start:
                # Can't move from start to I or E
                if to.startswith("I-") or to.startswith("E-"):
                    mask[vocab[to], vocab[from_]] = small
            else:
                if from_.startswith("B-"):
                    # Can't move from B to B, S, O, End, or Pad
                    if to.startswith(("B-", "S-", "O")) or to is end or to is pad:
                        mask[vocab[to], vocab[from_]] = small
                    # Can only move to matching I or E
                    elif to.startswith("I-") or to.startswith("E-"):
                        from_type = from_.split("-", maxsplit=1)[1]
                        to_type = to.split("-", maxsplit=1)[1]
                        if from_type != to_type:
                            mask[vocab[to], vocab[from_]] = small
                elif from_.startswith("I-"):
                    # Can't move from I to B, S, O, End or Pad
                    if to.startswith(("B-", "S-", "O")) or to is end or to is pad:
                        mask[vocab[to], vocab[from_]] = small
                    # Can only move to matching I or E
                    elif to.startswith("I-") or to.startswith("E-"):
                        from_type = from_.split("-", maxsplit=1)[1]
                        to_type = to.split("-", maxsplit=1)[1]
                        if from_type != to_type:
                            mask[vocab[to], vocab[from_]] = small
                elif from_.startswith(("E-", "S-", "O")):
                    # Can't move from E to I or E
                    # Can't move from S to I or E
                    # Can't move from O to I or E
                    if to.startswith("I-") or to.startswith("E-"):
                        mask[vocab[to], vocab[from_]] = small
    return mask


@export
def get_version(pkg):
    s = ".".join(pkg.__version__.split(".")[:2])
    return float(s)


@export
def revlut(lut: Dict) -> Dict:
    return {v: k for k, v in lut.items()}


@export
def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise Exception("Boolean value expected.")


@export
def is_sequence(x) -> bool:
    if isinstance(x, str):
        return False
    return isinstance(x, (collections.Sequence, collections.MappingView))


@export
def listify(x: Union[List[Any], Any]) -> List[Any]:
    """Take a scalar or list and make it a list iff not already a sequence or numpy array

    :param x: The input to convert
    :return: A list
    """
    if is_sequence(x) or isinstance(x, np.ndarray):
        return x
    return [x] if x is not None else []


@export
def read_json(filepath: str, default_value: Optional[Any] = None, strict: bool = False) -> Dict:
    """Read a JSON file in.  If no file is found and default value is set, return that instead.  Otherwise error

    :param filepath: str, A file to load
    :param default_value: If the file doesn't exist, return return this. Defaults to an empty dict.
    :param strict: bool, If true raise an error on file not found.

    :return: dict, The read JSON object
    """
    if not os.path.exists(filepath):
        if strict:
            raise IOError("No file {} found".format(filepath))
        return default_value if default_value is not None else {}
    with open(filepath) as f:
        return json.load(f)


@export
def read_yaml(filepath: str, default_value: Optional[Any] = None, strict: bool = False) -> Dict:
    """Read a YAML file in.  If no file is found and default value is set, return that instead.  Otherwise error

    :param filepath: str, A file to load
    :param default_value: If the file doesn't exist, return return this. Defaults to an empty dict.
    :param strict: bool, If true raise an error on file not found.

    :return: dict, The read yaml object
    """
    if not os.path.exists(filepath):
        if strict:
            raise IOError("No file {} found".format(filepath))
        return default_value if default_value is not None else {}
    with open(filepath) as f:
        import yaml
        from distutils.version import LooseVersion

        if LooseVersion(yaml.__version__) >= LooseVersion("5.1"):
            return yaml.load(f, Loader=yaml.FullLoader)
        return yaml.load(f)


@export
def read_config_file(config_file: str) -> Dict:
    """Read config file, optionally supports YAML, if dependency was already installed.  O.W. JSON plz

    :param config_file: (``str``) A path to a config file which should be a JSON file, or YAML if pyyaml is installed
    :return: (``dict``) An object
    """
    try:
        return read_yaml(config_file, strict=True)
    except:
        return read_json(config_file, strict=True)


@export
def validate_url(url: str) -> bool:
    regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    return re.match(regex, url) is not None



@export
def read_config_stream(config_stream) -> Dict:
    """Read config stream.  May be a path to a YAML or JSON file, or a str containing JSON or the name
    of an env variable, or even a JSON object directly

    :param config_stream:
    :return:
    """
    if isinstance(config_stream, (dict, list)) or config_stream is None:
        return config_stream
    if os.path.exists(config_stream) and os.path.isfile(config_stream):
        logger.info("Reading config file '{}'".format(config_stream))
        return read_config_file(config_stream)
    config = config_stream
    if config_stream.startswith("hub:"):
        vec = config_stream.split(":")
        version = vec[1]
        rest = ":".join(vec[2:])
        config_stream = f"http://raw.githubusercontent.com/mead-ml/hub/master/{version}/{rest}.yml"

    if config_stream.startswith("$"):
        logger.info("Reading config from '{}'".format(config_stream))
        config = os.getenv(config_stream[1:])
    
    else:
        if validate_url(config_stream):
            path_to_save, _ = urlretrieve(config_stream)
            return read_config_stream(path_to_save)
        else:
            logger.info("No file found '{}...', loading as string".format(config_stream[:12]))
    return json.loads(config)


@export
def write_sentence_conll(handle, sentence, gold, txt, idx2label):

    if len(txt) != len(sentence):
        txt = txt[: len(sentence)]

    try:
        for word, truth, guess in zip(txt, gold, sentence):
            handle.write("%s %s %s\n" % (word["text"], idx2label[truth], idx2label[guess]))
        handle.write("\n")
    except:
        logger.error("ERROR: Failed to write lines... closing file")
        handle.close()


@export
def ls_props(thing):
    """List all of the properties on some object

    :param thing: Some object
    :return: The list of properties
    """
    return [x for x in dir(thing) if isinstance(getattr(type(thing), x, None), property)]


@export
def idempotent_append(element: Any, data: List[Any]) -> List[Any]:
    """Append to a list if that element is not already in the list.

    :param element: The element to add to the list.
    :param data: `List` the list to add to.
    :returns: `List` the list with the element in it.
    """
    if element not in data:
        data.append(element)
    return data


@export
def parse_module_as_path(module_name):
    """Convert a path to a file to a format that it can be imported.

    :param module_name: The module as a path.
    :returns: `Tuple[str, str]` the module name (without a file ext) and the
        absolute path of the dir the file lives in (or '' if the module_name
        is just a filename).
    """
    module_dir, module_name = os.path.split(module_name)
    module_dir = os.path.realpath(os.path.expanduser(module_dir)) if module_dir else module_dir
    module_name, _ = os.path.splitext(module_name)
    return module_name, module_dir

@export
def fill_y(nc, yidx):
    """Convert a `B` sparse array to a dense one, to expand labels

    :param nc: (``int``) The number of labels
    :param yidx: The sparse array of the labels
    :return: A dense array
    """
    xidx = np.arange(0, yidx.shape[0], 1)
    dense = np.zeros((yidx.shape[0], nc), dtype=int)
    dense[xidx, yidx] = 1
    return dense


@export
@optional_params
def str_file(func, **kwargs):
    """A decorator to automatically open arguments that are files.

    If there are kwargs then they are name=mode. When the function is
    called if the argument name is a string then the file is opened with
    mode.

    If there are no kwargs then it is assumed the first argument is a
    file that should be opened as 'r'
    """
    possible_files = kwargs
    # We need to have the generator check out here so that the inner function is
    # either a generator (has a yield) or not. If we were to try to have this if
    # inside of the open_files then we would always return a generator.
    if inspect.isgeneratorfunction(func):

        @wraps(func)
        def open_files(*args, **kwargs):
            # If no arg names are given then assume the first arg is a file
            # you want to read from.
            if not possible_files:
                if isinstance(args[0], str):
                    with io.open(args[0], mode="r", encoding="utf-8") as f:
                        # Call the function with the file instead we need to
                        # yield from it until it is done other wise the file
                        # will be closed after the first yield.
                        for x in func(f, *args[1:], **kwargs):
                            yield x
                else:
                    for x in func(*args, **kwargs):
                        yield x
            else:
                # Otherwise we have multiple files we want to open
                to_close = []
                # Get a dict representation of what it will look like if we
                # call func with *args and **kwargs
                arg = inspect.getcallargs(func, *args, **kwargs)
                try:
                    for f, mode in possible_files.items():
                        if isinstance(arg[f], str):
                            # Replace strings with the opened files
                            arg[f] = io.open(arg[f], mode=mode, encoding=None if "b" in mode else "utf-8")
                            to_close.append(f)
                    # Call the function with the files instead
                    for x in func(**arg):
                        yield x
                finally:
                    # Make sure to close the files
                    for f in to_close:
                        arg[f].close()

    else:

        @wraps(func)
        def open_files(*args, **kwargs):
            # If no arg names are given then assume the first arg is a file
            # you want to read from.
            if not possible_files:
                if isinstance(args[0], str):
                    with io.open(args[0], mode="r", encoding="utf-8") as f:
                        # Call the function with the file instead
                        return func(f, *args[1:], **kwargs)
                else:
                    return func(*args, **kwargs)
            else:
                # Otherwise we have multiple files we want to open
                to_close = []
                # Get a dict representation of what it will look like if we
                # call func with *args and **kwargs
                arg = inspect.getcallargs(func, *args, **kwargs)
                try:
                    for f, mode in possible_files.items():
                        if isinstance(arg[f], str):
                            # Replace strings with the opened files
                            arg[f] = io.open(arg[f], mode=mode, encoding=None if "b" in mode else "utf-8")
                            to_close.append(f)
                    # Call the function with the files instead
                    return func(**arg)
                finally:
                    # Make sure to close the files
                    for f in to_close:
                        arg[f].close()

    return open_files


@export
@str_file(filepath="w")
def write_json(content, filepath):
    json.dump(content, filepath, indent=True)


@export
@str_file(filepath="w")
def write_yaml(content, filepath):
    import yaml

    yaml.dump(content, filepath, default_flow_style=False)


@export
def normalize_indices(xs, length):
    """Normalize negative indices into positive.

    :param xs: `List[int]` The indices.
    :param length: `int` The length of the thing to be indexed

    :returns: `List[int]` The indices converted to positive only.
    """
    return list(map(lambda x: length + x if x < 0 else x, xs))


@export
def convert_iob_to_bio(seq: List[str]) -> List[str]:
    """Convert a sequence of IOB tags to BIO tags.

    The difference between IOB and BIO (also called IOB2) is that in IOB
    the B- prefix is only used to separate two chunks of the same type
    while in BIO the B- prefix is used to start every chunk.

    :param seq: `List[str]` The list of IOB tags.

    :returns: `List[str] The list of BIO tags.
    """
    new = []
    prev = "O"
    for token in seq:
        # Only I- needs to be changed
        if token.startswith("I-"):
            # If last was O or last was a different type force B
            if prev == "O" or token[2:] != prev[2:]:
                token = "B-" + token[2:]
        new.append(token)
        prev = token
    return new


@export
def convert_bio_to_iob(seq: List[str]) -> List[str]:
    """Convert a sequence of BIO tags to IOB tags.

    The difference between BIO and IOB is that in IOB the B- prefix is only
    used to separate two chunks of the same type while in BIO the B- prefix
    starts every chunk. To convert we only need to look at the B- tokens.
    If they are following a chunk of the same type we leave it as a B-
    otherwise it converts it back to an I-

    :param seq: `List[str]` The list of BIO tags.

    :returns: `List[str]` The list of IOB tags.
    """
    new = []
    prev_ty = "O"
    for token in seq:
        ty = "O" if token == "O" else token[2:]
        # In IOB, `B-` is only needed if the previous type is the same as ours
        if token.startswith("B-"):
            # If we are different than the type before us convert to `I-`
            if prev_ty != ty:
                token = "I-" + ty
        new.append(token)
        prev_ty = ty
    return new


@export
def convert_bio_to_iobes(seq: List[str]) -> List[str]:
    """Convert a sequence of BIO tags to IOBES tags.

    The difference between BIO and IOBES tags is that in IOBES the end
    of a multi-token entity is marked with the E- prefix while in BIO
    it would end with an I- prefix.

    The other difference is that a single token entity in BIO is a
    just a B- whereas in IOBES it uses the special S- prefix.

    :param seq: `List[str]` The list of BIO tags.

    :returns: `List[str]` The list of IOBES tags.
    """
    new = []
    # Get tag bigrams with a fake O at the end so that the final real
    # token is processed correctly (Final I turned to E, etc.)
    for c, n in zip(seq, chain(seq[1:], ["O"])):
        if c.startswith("B-"):
            # If the next token is an I of the same class this is a
            # multi token span and should stay as B
            if n == c.replace("B-", "I-"):
                new.append(c)
            # If the next token is anything else this is a single token
            # span and should become a S
            else:
                new.append(c.replace("B-", "S-"))
        elif c.startswith("I-"):
            # If the next token is also an I of this type then we are
            # in the middle of a span and should stay I
            if n == c:
                new.append(c)
            # If next is anything else we are the end and thus E
            else:
                new.append(c.replace("I-", "E-"))
        # Pass O through
        else:
            new.append(c)
    return new


@export
def convert_iobes_to_bio(seq: List[str]) -> List[str]:
    """Convert a sequence of IOBES tags to BIO tags

    :param seq: `List[str]` The list of IOBES tags.

    :returns: `List[str]` The list of BIO tags.
    """
    # Use re over `.replace` to make sure it only acts on the beginning of the token.
    return list(map(lambda x: re.sub(r"^S-", "B-", re.sub(r"^E-", "I-", x)), seq))


@export
def convert_iob_to_iobes(seq: List[str]) -> List[str]:
    """Convert a sequence of IOB tags to IOBES tags.

    :param seq: `List[str]` The list of IOB tags.

    :returns: `List[str]` The list of IOBES tags.
    """
    return convert_bio_to_iobes(convert_iob_to_bio(seq))


@export
def convert_iobes_to_iob(seq: List[str]) -> List[str]:
    """Convert a sequence of IOBES tags to IOB tags.

    :param seq: `List[str]` The list of IOBES tags.

    :returns: `List[str]` The list of IOB tags.
    """
    return convert_bio_to_iob(convert_iobes_to_bio(seq))


@export
@str_file
def sniff_conll_file(f, delim=None, comment_pattern="#", doc_pattern="# begin doc"):
    """Figure out how many columns are in a conll file.

    :param file_name: `str` The name of the file.
    :param delim: `str` The token between columns in the file.

    :returns: `int` The number of columns in the file.
    """
    start = f.tell()
    for line in f:
        line = line.rstrip("\n")
        if line.startswith((comment_pattern, doc_pattern)):
            continue
        parts = line.split(delim)
        if len(parts) > 1:
            f.seek(start)
            return len(parts)

@export
def split_extensions(path: str, exts: Set[str]) -> Tuple[str, str]:
    all_ext = []
    path, ext = os.path.splitext(path)
    while ext in exts:
        all_ext.append(ext)
        path, ext = os.path.splitext(path)
    return path + ext, "".join(all_ext[::-1])

@export
def remove_extensions(path: str, exts: Set[str]) -> str:
    path, _ = split_extensions(path, exts)
    return path


@export
def remove_conll_extensions(path: str, exts: Set[str] = CONLL_EXTS) -> str:
    return remove_extensions(path, exts)

@export
def split_conll_extensions(path: str, exts: Set[str] = CONLL_EXTS) -> Tuple[str, str]:
    return split_extensions(path, exts)


@export
@str_file
def read_conll(f, doc_pattern=None, delim=None, metadata=False, allow_comments=False, comment_pattern="#"):
    """Read from a conll file.

    :param f: `str` The file to read from.
    :param doc_pattern: `str` A pattern that matches the line that signals the
        beginning of a new document in the conll file. When None just the
        conll sentences are returned.
    :param delim: `str` The token between columns in the file
    :param metadata: `bool` Should meta data (lines starting with `#` before a
        sentence) be returned with our sentences.
    :param allow_comments: `bool` Are comments (lines starting with `#`) allowed in the file.

    :returns: `Generator` The sentences or documents from the file.
    """
    if metadata and not allow_comments:
        raise ValueError(
            "You have metadata set to `True` but allow_comments set to `False` you can't extract "
            "metadata from a file that doesn't allow comments."
        )
    if doc_pattern is not None:
        if metadata:
            for x in read_conll_docs_md(f, doc_pattern, delim=delim, comment_pattern=comment_pattern):
                yield x
        else:
            for x in read_conll_docs(
                f, doc_pattern, delim=delim, allow_comments=allow_comments, comment_pattern=comment_pattern
            ):
                yield x
    else:
        if metadata:
            for x in read_conll_sentences_md(f, delim=delim, comment_pattern=comment_pattern):
                yield x
        else:
            for x in read_conll_sentences(
                f, delim=delim, allow_comments=allow_comments, comment_pattern=comment_pattern
            ):
                yield x


@export
@str_file
def read_conll_sentences(f, delim=None, allow_comments=True, comment_pattern="#"):
    """Read sentences from a conll file.

    :param f: `str` The file to read from.
    :param delim: `str` The token between columns in the file.

    Note:
        If you have a sentence where the first token is `#` it will get eaten by the
        metadata. If this happens you need to set `allow_comments=True` and not have
        comments in the file. If you have comments in the file and set this then
        they will show up in the sentences

    :returns: `Generator[List[List[str]]]` A list of rows representing a sentence.
    """
    sentence = []
    for line in f:
        line = line.rstrip()
        # Comments are not allowed in the middle of a sentence so if we find a line that
        # starts with # but we are in a sentence it must be a # as a token so we should
        # not skip it
        if allow_comments and not sentence and line.startswith(comment_pattern):
            continue

        # Blank lines signal the end of a sentence
        if len(line) == 0:
            # If we built a sentence yield it, this check allows multiple blank lines in a row
            if sentence:
                yield sentence
            # Reset the sentence
            sentence = []
            continue
        # This is a normal row, we split and take the tokens.
        sentence.append(line.split(delim))
    # If we have a sentence then the file didn't end with a new line, we yield the sentence
    # so we don't lose it
    if sentence:
        yield sentence


@export
@str_file
def read_conll_sentences_md(f, delim=None, comment_pattern="#"):
    """Read sentences from a conll file.

    :param f: `str` The file to read from.
    :param delim: `str` The token between columns in the file.

    Note:
        If there are document annotations in the conll file then they will show
        up in the meta data for what would be the first sentence of that doc

        If you have a sentence where the first token is `#` it will show up in the
        metadata. If this happens you'll need to update you comments to use a different
        comment pattern, something like `# comment:` I recommend having a space in
        you patten so it can't show up as a conll token

    :returns: `Generator[Tuple[List[List[str]], List[List[str]]]`
        The first element is the list or rows, the second is a list of comment
        lines that preceded that sentence in the file.
    """
    sentence, meta = [], []
    for line in f:
        line = line.rstrip()
        # Comments are not allowed in the middle of a sentence so if we find a line that
        # starts with # but we are in a sentence it must be a # as a token so we should
        # not skip it. If this is a comment we track it in our meta data list
        if not sentence and line.startswith(comment_pattern):
            meta.append(line)
            continue
        if len(line) == 0:
            if sentence:
                yield sentence, meta
            sentence, meta = [], []
            continue
        sentence.append(line.split(delim))
    if sentence:
        yield sentence, meta


@export
@str_file
def read_conll_docs(f, doc_pattern="# begin doc", delim=None, allow_comments=True, comment_pattern="#"):
    """Read sentences from a conll file.

    :param f: `str` The file to read from.
    :param doc_pattern: `str` The beginning of lines that represent new documents
    :param delim: `str` The token between columns in the file.

    Note:
        If you have a sentence where the first token is `#` it will show up in the
        metadata. If this happens you'll need to update you comments to use a different
        comment pattern, something like `# comment:` I recommend having a space in
        you patten so it can't show up as a conll token

    :returns: `Generator[List[List[List[str]]]]`
        A document which is a list of sentences.
    """
    doc, sentence = [], []
    for line in f:
        line = line.rstrip()
        if line.startswith(doc_pattern):
            if doc:
                if sentence:
                    doc.append(sentence)
                yield doc
                doc, sentence = [], []
            continue
        elif allow_comments and not sentence and line.startswith(comment_pattern):
            continue
        if len(line) == 0:
            if sentence:
                doc.append(sentence)
                sentence = []
            continue
        sentence.append(line.split(delim))
    if doc or sentence:
        if sentence:
            doc.append(sentence)
        yield doc


@export
@str_file
def read_conll_docs_md(f, doc_pattern="# begin doc", delim=None, comment_pattern="#"):
    """Read sentences from a conll file.

    :param f: `str` The file to read from.
    :param doc_pattern: `str` The beginning of lines that represent new documents
    :param delim: `str` The token between columns in the file.

    Note:
        If you have a sentence where the first token is `#` it will show up in the
        metadata. If this happens you'll need to update you comments to use a different
        comment pattern, something like `# comment:` I recommend having a space in
        you patten so it can't show up as a conll token

    :returns: `Generator[Tuple[List[List[List[str]]], List[str]  List[List[str]]]`
        The first element is a document, the second is a list of comments
        lines that preceded the document break (includes the document line)
        since the last sentence. The last is a list of comments for each
        list in the document.
    """
    doc, sentence, doc_meta, sent_meta, meta = [], [], [], [], []
    for line in f:
        line = line.rstrip()
        if line.startswith(doc_pattern):
            new_doc_meta = meta
            meta = []
            new_doc_meta.append(line)
            if doc:
                if sentence:
                    doc.append(sentence)
                    sentence = []
                yield doc, doc_meta, sent_meta
                doc, sentence, sent_meta = [], [], []
            doc_meta = new_doc_meta
            continue
        elif not sentence and line.startswith(comment_pattern):
            meta.append(line)
            continue
        if len(line) == 0:
            if sentence:
                doc.append(sentence)
                sent_meta.append(meta)
                sentence, meta = [], []
            continue
        sentence.append(line.split(delim))
    if doc or sentence:
        if sentence:
            doc.append(sentence)
            sent_meta.append(meta)
            meta = []
        yield doc, doc_meta, sent_meta


@export
@str_file(ifile="r", ofile="w")
def convert_conll_file(ifile, ofile, convert, fields=[-1], delim=None):
    """Convert the tagging scheme in a conll file.

    This function assumes the that columns that one wishes to convert are
    the right model columns.

    :param ifile: `str` The input file name.
    :param ofile: `str` The output file name.
    :param convert: `Callable(List[str]) -> List[str]` The function that
        transforms a sequence in one tag scheme to another scheme.
    :param fields: `List[int]` The columns to convert.
    :param delim: `str` The symbol that separates the columns.
    """
    output_delim = ' ' if delim is None else delim
    conll_length = sniff_conll_file(ifile, delim)
    fields = set(normalize_indices(fields, conll_length))
    for lines, md in read_conll_sentences_md(ifile, delim=delim):
        lines = zip(*(convert(l) if i in fields else l for i, l in enumerate(zip(*lines))))
        # Write out meta data
        if md:
            ofile.write("\n".join(md) + "\n")
        # Write out the lines
        ofile.write("\n".join(output_delim.join(l).rstrip() for l in lines) + "\n\n")


@export
@str_file(ifile="r", ofile="w")
def convert_iob_conll_to_bio(ifile, ofile, fields=[-1], delim=None):
    """Convert a conll file from iob to bio."""
    convert_conll_file(ifile, ofile, convert_iob_to_bio, fields, delim)


@export
@str_file(ifile="r", ofile="w")
def convert_bio_conll_to_iob(ifile, ofile, fields=[-1], delim=None):
    """Convert a conll file from bio to iob."""
    convert_conll_file(ifile, ofile, convert_bio_to_iob, fields, delim)


@export
@str_file(ifile="r", ofile="w")
def convert_iob_conll_to_iobes(ifile, ofile, fields=[-1], delim=None):
    """Convert a conll file from iob to iobes."""
    convert_conll_file(ifile, ofile, convert_iob_to_iobes, fields, delim)


@export
@str_file(ifile="r", ofile="w")
def convert_iobes_conll_to_iob(ifile, ofile, fields=[-1], delim=None):
    """Convert a conll file from iob to iobes."""
    convert_conll_file(ifile, ofile, convert_iobes_to_iob, fields, delim)


@export
@str_file(ifile="r", ofile="w")
def convert_bio_conll_to_iobes(ifile, ofile, fields=[-1], delim=None):
    """Convert a conll file from bio to iobes."""
    convert_conll_file(ifile, ofile, convert_bio_to_iobes, fields, delim)


@export
@str_file(ifile="r", ofile="w")
def convert_iobes_conll_to_bio(ifile, ofile, fields=[-1], delim=None):
    """Convert a conll file from iobes to bio. Useful for formatting output to use `conlleval.pl`."""
    convert_conll_file(ifile, ofile, convert_iobes_to_bio, fields, delim)


@export
def to_spans(
    sequence: List[int], lut: Dict[int, str], span_type: str, verbose: bool = False, delim: str = "@"
) -> List[str]:
    """Turn a sequence into a list of chunks.

    :param sequence: `List[int]` The tag sequence.
    :param lut: `Dict[int] -> str` A mapping for integers to tag names.
    :param span_type: `str` The tagging scheme.
    :param verbose: `bool` Should we output warning on illegal transitions.
    :param delim: `str` The symbol the separates output chunks from their indices.

    :returns: `List[str]` The list of entities in the order they appear. The
        entities are in the form {chunk_type}{delim}{index}{delim}{index}...
        for example LOC@3@4@5 means a Location chunk was at indices 3, 4, and 5
        in the original sequence.
    """
    sequence = [lut[y] for y in sequence]
    return to_chunks(sequence, span_type, verbose, delim)

@export
def to_chunks(sequence: List[str], span_type: str, verbose: bool = False, delim: str = "@") -> List[str]:
    """Turn a sequence of tags into a list of chunks.

    :param sequence: `List[str]` The tag sequence.
    :param span_type: `str` The tagging scheme.
    :param verbose: `bool` Should we output warning on illegal transitions.
    :param delim: `str` The symbol the separates output chunks from their indices.

    :returns: `List[str]` The list of entities in the order they appear. The
        entities are in the form {chunk_type}{delim}{index}{delim}{index}...
        for example LOC@3@4@5 means a Location chunk was at indices 3, 4, and 5
        in the original sequence.
    """
    span_type = span_type.lower()
    if span_type == "iobes":
        return to_chunks_iobes(sequence, verbose, delim)
    if span_type == "token":
        # For token level tasks we force each token to be it's own span
        # Normally it is fine to pass token level annotations through the iob
        # span code to produce spans of size 1 but if your data has a label of `O`
        # like twpos does then you will lose tokens and mess up the calculations
        return [f"{s}@{i}" for i, s in enumerate(sequence)]

    strict_iob2 = span_type in ("iob2", "bio")
    iobtype = 2 if strict_iob2 else 1
    chunks = []
    current = None
    for i, label in enumerate(sequence):
        if not label.startswith("I-") and not label == "O":
            if current is not None:
                chunks.append(delim.join(current))
            current = [label.replace("B-", ""), "%d" % i]
        elif label.startswith("I-"):
            if current is not None:
                base = label.replace("I-", "")
                if base == current[0]:
                    current.append("%d" % i)
                else:
                    chunks.append(delim.join(current))
                    if iobtype == 2 and verbose:
                        logger.warning("Warning: I doesn't agree with previous B/I @ %d" % i)

                    current = [base, "%d" % i]
            else:
                current = [label.replace("I-", ""), "%d" % i]
                if iobtype == 2 and verbose:
                    logger.warning("Warning: I without previous chunk @ %d" % i)
        else:
            if current is not None:
                chunks.append(delim.join(current))
            current = None
    if current is not None:
        chunks.append(delim.join(current))
    return chunks


@export
def to_chunks_iobes(sequence: List[str], verbose: bool = False, delim: str = "@") -> List[str]:
    """Turn a sequence of IOBES tags into a list of chunks.

    :param sequence: `List[str]` The tag sequence.
    :param verbose: `bool` Should we output warning on illegal transitions.
    :param delim: `str` The symbol the separates output chunks from their indices.

    :returns: `List[str]` The list of entities in the order they appear. The
        entities are in the form {chunk_type}{delim}{index}{delim}{index}...
        for example LOC@3@4@5 means a Location chunk was at indices 3, 4, and 5
        in the original sequence.
    """
    chunks = []
    current = None
    for i, label in enumerate(sequence):
        # This indicates a multi-word chunk start
        if label.startswith("B-"):
            # Flush existing chunk
            if current is not None:
                chunks.append(delim.join(current))
            # Create a new chunk
            current = [label.replace("B-", ""), "%d" % i]
            if verbose:
                # Look ahead to make sure this `B-Y` shouldn't be a `S-Y`
                # We only check for `B`, `S` and `O` because a mismatched `I` or `E`
                # will already warn looking backwards
                if i < len(sequence) - 1:
                    nxt = sequence[i + 1]
                    if nxt == "O" or nxt.startswith(("B", "S")):
                        logger.warning("Warning: Single B token chunk @ %d", i)
                elif i == len(sequence) - 1:
                    logger.warning("Warning: B as final token")
        # This indicates a single word chunk
        elif label.startswith("S-"):
            # Flush existing chunk, and since this is self-contained, we will clear current
            if current is not None:
                chunks.append(delim.join(current))
                current = None
            base = label.replace("S-", "")
            # Write this right into the chunks since self-contained
            chunks.append(delim.join([base, "%d" % i]))
        # Indicates we are inside of a chunk already
        elif label.startswith("I-"):
            # This should always be the case!
            if current is not None:
                base = label.replace("I-", "")
                if base == current[0]:
                    current.append("%d" % i)
                else:
                    # Doesn't match previous entity, flush the old one and start a new one
                    chunks.append(delim.join(current))
                    if verbose:
                        logger.warning("Warning: I doesn't agree with previous B/I @ %d" % i)
                    current = [base, "%d" % i]
            else:
                if verbose:
                    logger.warning("Warning: I without previous chunk @ %d" % i)
                current = [label.replace("I-", ""), "%d" % i]
            if verbose and i == len(sequence) - 1:
                logger.warning("Warning: I as final token")
        # We are at the end of a chunk, so flush current
        elif label.startswith("E-"):
            # Flush current chunk
            if current is not None:
                base = label.replace("E-", "")
                if base == current[0]:
                    current.append("%d" % i)
                    chunks.append(delim.join(current))
                    current = None
                else:
                    chunks.append(delim.join(current))
                    if verbose:
                        logger.warning("Warning: E doesn't agree with previous B/I @ %d", i)
                    current = [base, "%d" % i]
                    chunks.append(delim.join(current))
                    current = None
            # This should never happen
            else:
                current = [label.replace("E-", ""), "%d" % i]
                if verbose:
                    logger.warning("Warning: E without previous chunk @ %d" % i)
                chunks.append(delim.join(current))
                current = None
        # Outside
        else:
            if current is not None:
                chunks.append(delim.join(current))
            current = None
    # If something is left, flush
    if current is not None:
        chunks.append(delim.join(current))
    return chunks


@export
def span_f1(golds: List[Set[str]], preds: List[Set[str]]) -> float:
    """Calculate Span level F1 score.

    :param golds: `List[set[str]]` The list of the set of gold chunks.
    :param preds: `List[set[str]]` The list of the set of predicted chunks.

    :returns: `float` The f1 score.
    """
    overlap = sum(len(g & p) for g, p in zip(golds, preds))
    gold_total = sum(len(g) for g in golds)
    pred_total = sum(len(p) for p in preds)
    return f_score(overlap, gold_total, pred_total)


@export
def per_entity_f1(golds: List[Set[str]], preds: List[Set[str]], delim: str = "@") -> Dict[str, float]:
    """Calculate Span level F1 with break downs per entity type.

    :param golds: `List[set[str]]` The list of the set of gold chunks.
    :param preds: `List[set[str]]` The list of the set of predicted chunks.
    :param delim: `str` The symbol that separates an entity from its indices.

    :returns: `dict` The metrics at a global level and fine grained entity
        level performance.

    Note:
        This function returns most of the metrics needed for the
        `conlleval_output`. `acc` and `tokens` (the token level accuracy
        and the number of tokens respectively) need to be added.
    """
    metrics = {}
    overlap = Counter()
    gold_total = Counter()
    pred_total = Counter()
    types = set()
    for g, p in zip(golds, preds):
        overlaps = g & p
        overlap["total"] += len(overlaps)
        gold_total["total"] += len(g)
        pred_total["total"] += len(p)
        for o in overlaps:
            ent = o.split(delim)[0]
            overlap[ent] += 1
            types.add(ent)
        for o in g:
            ent = o.split(delim)[0]
            gold_total[ent] += 1
            types.add(ent)
        for o in p:
            ent = o.split(delim)[0]
            pred_total[ent] += 1
            types.add(ent)
    metrics["overlap"] = overlap["total"]
    metrics["gold_total"] = gold_total["total"]
    metrics["pred_total"] = pred_total["total"]
    metrics["precision"] = precision(overlap["total"], pred_total["total"]) * 100
    metrics["recall"] = recall(overlap["total"], gold_total["total"]) * 100
    metrics["f1"] = f_score(overlap["total"], gold_total["total"], pred_total["total"]) * 100
    metrics["types"] = []
    for t in sorted(types):
        metrics["types"].append(
            {
                "ent": t,
                "precision": precision(overlap[t], pred_total[t]) * 100,
                "recall": recall(overlap[t], gold_total[t]) * 100,
                "f1": f_score(overlap[t], gold_total[t], pred_total[t]) * 100,
                "count": pred_total[t],
            }
        )
    return metrics


@export
def conlleval_output(results: Dict[str, Union[float, int]]) -> str:
    """Create conlleval formated output.

    :param results: `dict` The metrics. results should have the following keys.
        tokens: `int` The total number of tokens processed.
        acc: `float` The token level accuracy.
        gold_total: `int` The total number of gold entities.
        pred_total: `int` The total number of predicted entities.
        overlap: `int` The number of exact match entites.
        precision: `float` The precision of all entities.
        recall: `float` The recall of all entities.
        f1: `float` The f1 score of all entities.
        types: `List[dict]` A list of metrics for each entity type. Keys should include:
            ent: `str` The name of the entity.
            precision: `float` The precision of this entity type.
            recall: `float` The recall of this this entity type.
            f1: `float` The f1 score of this entity type.
            count: `int` The number of predicted entities of this type.

    :returns: `str` The formatted string ready for printing.

    Note:
        Both the metrics in the results dict and acc are expected to already be
        multiplied by 100. The result won't look correct and a lot of the
        metric will be cut off if they are not.

        Metrics per type are output in the order they appear in the list.
        conlleval.pl outputs the types in sorted order. To match this the list
        in `results['types'] should be sorted.
    """
    s = (
        "processed {tokens} tokens with {gold_total} phrases; found: {pred_total} phrases; correct: {overlap}.\n"
        "accuracy: {acc:>{length}.2f}%; precision: {precision:>6.2f}%; recall: {recall:>6.2f}%; FB1: {f1:>6.2f}\n"
    )
    t = []
    longest_ent = max(len(max(results["types"], key=lambda x: len(x["ent"]))["ent"]), 17)
    for type_metric in results["types"]:
        t.append(
            "{ent:>{longest_ent}}: precision: {precision:>6.2f}%; recall: {recall:>6.2f}%; FB1: {f1:>6.2f}  {count}".format(
                longest_ent=longest_ent, **type_metric
            )
        )
    s = s + "\n".join(t)
    s = s.format(length=longest_ent - 11, **results)
    return s


@export
def precision(overlap_count: int, guess_count: int) -> float:
    """Compute the precision in a zero safe way.

    :param overlap_count: `int` The number of true positives.
    :param guess_count: `int` The number of predicted positives (tp + fp)

    :returns: `float` The precision.
    """
    if guess_count == 0:
        return 0.0
    return 0.0 if guess_count == 0 else overlap_count / float(guess_count)


@export
def recall(overlap_count: int, gold_count: int) -> float:
    """Compute the recall in a zero safe way.

    :param overlap_count: `int` The number of true positives.
    :param gold_count: `int` The number of gold positives (tp + fn)

    :returns: `float` The recall.
    """
    return 0.0 if gold_count == 0 else overlap_count / float(gold_count)


@export
def f_score(overlap_count: int, gold_count: int, guess_count: int, f: int = 1) -> float:
    """Compute the f1 score.

    :param overlap_count: `int` The number of true positives.
    :param gold_count: `int` The number of gold positives (tp + fn)
    :param guess_count: `int` The number of predicted positives (tp + fp)
    :param f: `int` The beta term to weight precision vs recall.

    :returns: `float` The f score
    """
    beta_sq = f * f
    if guess_count == 0:
        return 0.0
    p = precision(overlap_count, guess_count)
    r = recall(overlap_count, gold_count)
    if p == 0.0 or r == 0.0:
        return 0.0
    f = (1.0 + beta_sq) * (p * r) / (beta_sq * p + r)
    return f


@export
def get_env_gpus() -> List[str]:
    return os.getenv("CUDA_VISIBLE_DEVICES", os.getenv("NVIDIA_VISIBLE_DEVICES", os.getenv("NV_GPU", "0"))).split(",")


@export
def ngrams(sentence: List[str], filtsz: int = 3, joiner: str = "@@") -> List[str]:
    """Generate ngrams over a sentence

    :param sentence: (`List[str]`) Some tokens
    :param filtsz: The ngram width
    :param joiner: A string to join ngrams
    :return: (`List[str]`) A list of ngrams
    """
    chunks = []
    nt = len(sentence)
    for i in range(nt - filtsz + 1):
        chunk = sentence[i : i + filtsz]
        chunks += [joiner.join(chunk)]
    return chunks


@export
@str_file
def read_label_first_data(f):
    """Read data from a file where the first token in the label and each line is a single example.

    :param f: `Union[str, IO]` The file to read from.
    :return: `Tuple[List[str], List[List[str]]]` The labels and text
    """
    labels, texts = [], []
    for line in f:
        line = line.rstrip()
        if line:
            label, text = line.split(maxsplit=1)
            labels.append(label)
            texts.append(text.split())
    return labels, texts


@export
@str_file(w="w")
def write_label_first_data(w, labels, texts):
    """Read data to a file where the first token in the label and each line is a single example.

    :param w: `Union[str, IO]` The file to write the results in
    :param labels: `List[str]` The labels for the examples
    :param texts: `List[List[str]]` The text examples
    """
    w.write("\n".join(" ".join(chain((l,), t)) for l, t in zip(labels, texts)))


class MN:
    GZIP = b"1f8b"
    TAR = b"7573746172"
    TAR_START = 257
    ZIP = b"504b0304"


def check_mn(b: bytes, mn: bytes = None, start: int = 0) -> bool:
    if hexlify(b[start : start + 20])[: len(mn)] == mn:
        return True
    return False


def int2byte(x: int) -> bytes:
    """Takes an int in the 8-bit range and returns a single-character byte"""
    return bytes((x,))


_text_characters: bytes = (b"".join(int2byte(i) for i in range(32, 127)) + b"\n\r\t\f\b")


# Borrowed from: https://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python
def is_text_file(block: bytes):
    """ Uses heuristics to guess whether the given file is text or binary,
        by reading a single block of bytes from the file.
        If more than 30% of the chars in the block are non-text, or there
        are NUL ('\x00') bytes in the block, assume this is a binary file.
    """
    if b"\x00" in block:
        # Files with null bytes are binary
        return False
    elif not block:
        # An empty file is considered a valid text file
        return True

    # Use translate's 'deletechars' argument to efficiently remove all
    # occurrences of _text_characters from the block
    nontext = block.translate(None, _text_characters)
    return float(len(nontext)) / len(block) <= 0.30


check_gzip = partial(check_mn, mn=MN.GZIP)
check_tar = partial(check_mn, mn=MN.TAR, start=MN.TAR_START)
check_zip = partial(check_mn, mn=MN.ZIP)


class RE(object):
    HTML = re.compile(b"(<!doctype html>|<html.*?>)")
    BIN = re.compile(b"\d+? \d+?$", re.MULTILINE)


def check_re(b: bytes, regex: Pattern = None) -> bool:
    return True if regex.match(b) else False


check_html = partial(check_re, regex=RE.HTML)


@export
def mime_type(file_name: str) -> str:
    b = open(file_name, "rb").read(1024)
    if check_gzip(b):
        return "application/gzip"
    if check_tar(b):
        return "application/x-tar"
    if check_zip(b):
        return "application/zip"
    if check_html(b):
        return "text/html"
    if is_text_file(b):
        return "text/plain"
    return "application/w2v"


@export
def to_numpy(x):
    if isinstance(x, np.ndarray):
        return x
    try:
        import torch
        if isinstance(x, torch.Tensor):
            x = x.detach()
            x = x.to('cpu')
    except ImportError:
        pass
    return x.numpy()


# https://stackoverflow.com/questions/5909873/how-can-i-pretty-print-ascii-tables-with-python
@export
def print_table(rows: collections.namedtuple, columns: Optional[Set[str]] = None) -> None:
    """Pretty print a table

    :param rows: Some rows
    :param columns: A set of columns to include in the output.
    """
    columns = columns if columns is not None else set(rows[0]._fields)
    fields = [f for f in rows[0]._fields if f in columns]
    if len(rows) > 1:
        lens = [
            len(str(max(chain((getattr(x, field) for x in rows), [field]), key=lambda x: len(str(x)))))
            for field in fields
        ]
        formats = []
        hformats = []
        for i, field in enumerate(fields):
            if isinstance(getattr(rows[0], field), int):
                formats.append("%%%dd" % lens[i])
            else:
                formats.append("%%-%ds" % lens[i])
            hformats.append("%%-%ds" % lens[i])
        pattern = " | ".join(formats)
        hpattern = " | ".join(hformats)
        separator = "-+-".join(['-' * n for n in lens])
        print(hpattern % tuple(fields))
        print(separator)
        for line in rows:
            print(pattern % tuple(getattr(line, field) for field in fields))
    elif len(rows) == 1:
        row = rows[0]
        hwidth = len(max(fields, key=lambda x: len(x)))
        for field in fields:
            print("%*s = %s" % (hwidth, field, getattr(row, field)))




r"""Multi-bleu in python matching Moses http://www.statmt.org/moses/?n=Moses.SupportTools#ntoc5

Reference files and prediction files must be sentence aligned. To use multiple
references use multiple reference files (all sentence aligned)

All text should already be tokenized (so that calling `.split()` on the line produces correct tokens)

Slight differences between this and multi-ble.pl:

 * multi-bleu.pl crashes when the hypothesis corpus is completely empty. We don't want training to crash
   so we set the brevity penalty to `0` in this case. This makes sense because as the reference corpus grows
   arbitrary large the length ratio (gold / pred) approaches infinity and the limit of e^(1 - x) as
   x -> infinity is 0. When the hypothesis corpus is we also return a np.nan for the length ratio for a
   visual cue that the lengths were weird when using the cli tool.

 * If your reference file doesn't have a newline at the end of the file but your hypothesis does
   then multi-blue.pl will give include this newline as an empty example and your score will
   be lower than it should be. This code will ignore that line and give the correct score.

   For example:
   ```
       $ diff ref pred
       1000c1000
       < 52 95 73 83 93 87 88 81 93
       \ No newline at end of file
       ---
       > 52 95 73 83 93 87 88 81 93

       $ diff ref pred_no_newline
       $

       $ bleu ref < pred
       BLEU = 100.00, 100.0/100.0/100.0/100.0 (BP=1.000, ratio=1.000, hyp_len=7533, ref_len=7533)

       $ bleu ref < pred_no_newline
       BLEU = 100.00, 100.0/100.0/100.0/100.0 (BP=1.000, ratio=1.000, hyp_len=7533, ref_len=7533)

       $ perl multi-bleu.pl ref < pred
       BLEU = 99.98, 100.0/100.0/100.0/100.0 (BP=1.000, ratio=1.000, hyp_len=7533, ref_len=7533)

       $ perl multi-bleu.pl ref < pred_no_newline
       BLEU = 100.00, 100.0/100.0/100.0/100.0 (BP=1.000, ratio=1.000, hyp_len=7533, ref_len=7533)
    ```
"""
import sys
import argparse
from operator import or_
from itertools import chain
from functools import reduce
from collections import Counter
from typing import List, Tuple, Union, Iterable, TextIO, Counter as CounterType, NamedTuple
import numpy as np


__all__ = ["Bleu", "bleu"]


Hypothesis = List[str]
# This is [B, T] where B is the number of hypotheses and T is the number of words in that hypothesis.
HypothesisCorpus = List[Hypothesis]

Reference = List[str]
References = List[Reference]
# This is [B, R, T] where B is the number of references, R is the number of gold references we have
# for a particular example, and T is the number of words in that reference.
ReferenceCorpus = List[References]


class Bleu(NamedTuple):
    """A collection of the information returned from the bleu calculation"""

    bleu: float
    precision: np.ndarray
    brevity_penalty: float
    length_ratio: float
    pred_length: int
    gold_length: int


def n_grams(tokens: List[str], n: Union[int, Tuple[int]]) -> Iterable[Tuple[str]]:
    """Create a list for n grams for each value up to and including n.

    :param tokens: The sequence to create n_grams on.
    :param n: If an int the largest n_gram to create otherwise the size of
        n_grams to create.

    :returns: The n_grams ordered by size.
    """
    n = range(n + 1) if isinstance(n, int) else n
    return chain(*(zip(*[tokens[i:] for i in range(n_)]) for n_ in n))


def count_n_grams(tokens: List[str], n: Union[int, Tuple[int]]) -> CounterType[Tuple[str]]:
    """Count the n_grams in a sequence.

    :param tokens: The sequence to create n_grams on.
    :param n: If an int the largest n_gram to create otherwise the size of
        n_grams to create.

    :returns: The counts of each n_gram.
    """
    return Counter(n_grams(tokens, n))


def find_closest(pred_len: int, golds: References) -> int:
    """Find the gold sentence that has the most similar length to pred.

    Note:
        When there are multiple sentence with the same difference in length
        to the pred length the shortest one is selected.

    :param pred_len: The length of the predicted sentence.
    :param golds: The gold sentences.

    :returns: The length of the selected gold sentence.
    """
    best_diff = sys.maxsize
    best_len = sys.maxsize
    for gold in golds:
        gold_len = len(gold)
        diff = abs(pred_len - gold_len)
        if diff < best_diff:
            best_diff = diff
            best_len = gold_len
        elif diff == best_diff:
            best_len = gold_len if gold_len < best_len else best_len
    return best_len


def corpora_lengths(preds: HypothesisCorpus, golds: ReferenceCorpus) -> Tuple[int, int]:
    """Calculate the length of the two corpora.

    The length of the pred corpus is just the sum of lengths, the length
    of the gold corpus is the sum of the lengths of a single gold selected
    from a set of golds. This single gold is selected to be the length that
    is closest to the pred lengths (in the event of ties select the shorter
    gold sentence).

    :param preds: A list of sentences generated by the model.
    :param golds: A list of gold references. A gold reference can have multiple gold sentences

    :returns: The length of the predicted corpus and the length of the gold corpus.
    """
    pred_len = 0
    gold_len = 0
    for pred, gold in zip(preds, golds):
        pred_len += len(pred)
        gold_len += find_closest(len(pred), gold)
    return pred_len, gold_len


def max_gold_n_gram_counts(golds: References, n: int) -> CounterType[Tuple[str]]:
    """Find the maximum n gram count for the gold sequences.

    :param golds: The gold sentences.
    :param n: The max size of n-grams we are using.

    :returns: The maximum count for any n-gram that appears in a gold sentence.
    """
    # or_ is the union of counters so result is max count for each n gram
    # that appears in the any of the gold sentences.
    return reduce(or_, map(lambda g: count_n_grams(g, n), golds))


def count_matches(
    pred_counts: CounterType[Tuple[str]], gold_counts: CounterType[Tuple[str]], matches: np.ndarray
) -> np.ndarray:
    """Aggregate the number of matches for each n gram length.

    :param pred_counts: The counts for n grams found in the predicted sentence.
    :param gold_counts: The max counts for n grams found in the gold sentences.
    :param matches: The number of matches found so far grouped by n-gram size.

    :returns: The number of matches so far grouped by n-gram size.
    """
    # & is the intersection of counters, selecting the min of each n gram
    # that appears in both of the counters.
    overlap = pred_counts & gold_counts
    for n_gram, count in overlap.items():
        matches[len(n_gram) - 1] += count
    return matches


def count_possible(pred: Hypothesis, total: np.ndarray) -> np.ndarray:
    """Count the total number of possible matches.

    We recalculate the total possible rather than use the counts calculated
    before because the counts aren't binned by length.

    :param pred: The predicted sentence.
    :param total: The current total number of possible matches grouped by size.

    :returns: The new number of possible n-gram matches of far grouped by size.
    """
    for n in range(len(total)):
        # Possible number of n_grams for a sequence is the length of the sequence
        # minus the length of the n_gram plus 1. In this case n is already
        # one less than length of n_gram so just subtract that.
        total_grams = len(pred) - n
        if total_grams > 0:
            total[n] += total_grams
    return total


def geometric_mean(precision: np.ndarray) -> float:
    """Calculate the geometric mean of the precision.

    Geometric mean is the nth root of the product of n numbers. This
    can be expressed as the e^(arithmetic mean of the logs). The geometric
    mean only applies to positive number and any 0 is the values makes the
    answer trivially zero to checkout for that first to avoid log(0).

    :param precision: The precision for each n-gram size.

    :returns: The geometric_mean of the values.
    """
    if np.min(precision) <= 0:
        return 0.0
    return np.exp(np.mean(np.log(precision))).item()


def brevity_penalty(pred_len: int, gold_len: int) -> Tuple[float, float]:
    """Calculate the brevity penalty.

    Note:
        multi-bleu.pl crashes when the hypothesis corpus is completely empty.
        We don't want training to crash then there are no hypotheses so we set
        the brevity penalty to `0` because as the reference corpus grows arbitrary
        large the length ration (gold / pred) approaches infinity and the limit of
        e^(1 - x) as x -> infinity is 0.

        We also return a np.nan for the length ratio for a visual cue that the
        lengths were weird when using the cli tool.

    :param pred_len: The length of the model prediction corpus
    :param gold_len: The length of the gold corpus

    :returns: The brevity penalty and the ratio of predicted length to gold length.
    """
    if pred_len == 0:
        return 0, np.nan
    ratio = pred_len / float(gold_len)
    # If ratio is <= 1.0 then pred_len <= gold_len so penalty applies.
    # Penalty is defined as e^(1 - (gold / pred)). (1 / (p / g)) = (g / p)
    bp = np.exp(1 - (1.0 / ratio)).item() if ratio <= 1.0 else 1.0
    return bp, ratio


def bleu(preds: HypothesisCorpus, golds: ReferenceCorpus, n: int = 4) -> Bleu:
    """Calculate BLEU score

    This implementation is designed to match the output of `multi-bleu,pl` from
    http://www.statmt.org/moses/?n=Moses.SupportTools#ntoc5

    :param preds: A list of sentences generated by the model.
    :param golds: A list of gold references where each reference can contain multiple sentences.
    :param n: The max size n-gram to use.

    :returns: (
        bleu score,
        precision per n_gram size,
        brevity penalty,
        ration of prediction length to reference lengths,
        length of the predictions,
        length of the references
    )
    """
    matches = np.zeros(n)
    total = np.zeros(n)
    pred_len, gold_len = corpora_lengths(preds, golds)
    for pred, gold in zip(preds, golds):
        max_gold_counts = max_gold_n_gram_counts(gold, n)
        pred_counts = count_n_grams(pred, n)
        matches = count_matches(pred_counts, max_gold_counts, matches)
        total = count_possible(pred, total)

    precision = np.array([matches[i] / float(total[i]) if total[i] > 0 else 0.0 for i in range(n)])
    geo_mean = geometric_mean(precision)
    bp, len_ratio = brevity_penalty(pred_len, gold_len)
    b = geo_mean * bp * 100
    return Bleu(b, precision * 100, bp, len_ratio, pred_len, gold_len)


def _read_references(reference_files: List[str], lc: bool) -> ReferenceCorpus:
    """Read from multiple reference files.

    :param reference_file: A list of reference files to read from.
    :param lc: Should the input be lowercased?

    :return: The reference corpus.
    """
    references: List[HypothesisCorpus] = []
    for ref in reference_files:
        with open(ref) as f:
            references.append(_read_lines(f, lc))
    return list(zip(*references))


# Slightly strange file reader that read from pre opened files to facilitate
# reading the predictions from stdin. Won't be used elsewhere.
def _read_lines(f: TextIO, lc: bool) -> HypothesisCorpus:
    """Read from a file that is open.

    The expected file format is each line contains a single example and each example
    is a list of tokens joined on whitespace

    :param f: The file to read from.
    :param lc: Should the input be lowercased?

    :return: The list of examples
    """
    if lc:
        return list(map(lambda x: list(map(lambda y: y.lower(), x.split())), f))
    return list(map(lambda x: x.split(), f))


def main():
    """Use as a cli tools like multibleu.pl"""
    usage = "%(prog)s [-lc] [-n] reference < hypothesis\nReads the references from reference or reference0, reference1, ...\n"
    parser = argparse.ArgumentParser(description="Calculate Bleu score.", usage=usage)
    parser.add_argument("-lc", help="lowercase the input", action="store_true")
    parser.add_argument("-n", type=int, default=4, help="The number of ngrams to use.")
    parser.add_argument("reference", nargs="+")
    args = parser.parse_args()

    golds: ReferenceCorpus = _read_references(args.reference, args.lc)
    preds: HypothesisCorpus = _read_lines(sys.stdin, args.lc)

    b = bleu(preds, golds, args.n)
    precision_str = "/".join(["{:.1f}"] * len(b.precision)).format(*b.precision)

    print(
        f"BLEU = {b.bleu:.2f}, {precision_str} (BP={b.brevity_penalty:.3f}, ratio={b.length_ratio:.3f}, hyp_len={b.pred_length}, ref_len={b.gold_length})"
    )


if __name__ == "__main__":
    main()



import re
import sys
import six
from eight_mile.utils import exporter


__all__ = []
export = exporter(__all__)


@export
class Progress(object):
    """Progress hook

    Provide interface for progress updates
    """

    def __init__(self):
        pass

    def update(self, step=1):
        """Update progress by number of `steps`

        :param step: The number of tasks completed
        :return:
        """
        pass

    def done(self):
        """Triggered when all tasks are completed

        :return:
        """
        pass

    def __call__(self, real_iter):
        return self.__iter__(real_iter)

    def __iter__(self, real_iter):
        for x in real_iter:
            yield x
            self.update()
        self.done()


@export
class ProgressBarJupyter(Progress):
    """Simple Jupyter progress bar

    Writes a progress bar to an ipython widget

    """

    def __init__(self, total):
        super(ProgressBarJupyter, self).__init__()
        from ipywidgets import FloatProgress
        from IPython.display import display

        self.progress = FloatProgress(min=0, max=total)
        display(self.progress)

    def update(self, step=1):
        self.progress.value += step

    def done(self):
        """Close the widget
        """
        self.progress.close()


# Modifed from here
# http://stackoverflow.com/questions/3160699/python-progress-bar#3160819
@export
class ProgressBarTerminal(Progress):
    """Simple terminal-based progress bar

    Writes a progress bar to the terminal, using a designated `symbol` (which defaults to `=`)

    """

    DEFAULT = "Progress: %(bar)s %(percent)3d%%"
    FULL = "%(bar)s %(current)d/%(total)d (%(percent)3d%%) %(remaining)d to go"

    def __init__(self, total, width=40, fmt=DEFAULT, symbol="="):
        super(ProgressBarTerminal, self).__init__()
        assert len(symbol) == 1

        self.total = total
        self.width = width
        self.symbol = symbol
        self.fmt = re.sub(r"(?P<name>%\(.+?\))d", r"\g<name>%dd" % len(str(total)), fmt)

        self.current = 0
        self.print_ = six.print_ if sys.stdout.isatty() else lambda *args, **kwargs: None
        # Force initial print of pg for when steps are long.
        self.update(step=0)

    def update(self, step=1):
        """Update progress bar by number of `steps`

        :param step: The number of tasks completed
        :return:
        """
        self.current += step
        percent = self.current / float(self.total)
        size = int(self.width * percent)
        remaining = self.total - self.current
        bar = "[" + self.symbol * size + " " * (self.width - size) + "]"

        args = {
            "total": self.total,
            "bar": bar,
            "current": self.current,
            "percent": percent * 100,
            "remaining": remaining,
        }
        self.print_("\r" + self.fmt % args, end="")

    def done(self):
        """All tasks are finished, complete the progress bar

        :return:
        """
        self.current = self.total
        self.update(step=0)
        self.print_("")



import numpy as np
from typing import Dict, List
from eight_mile.tf.layers import TransformerEncoderStack, TransformerEncoder, MultiHeadedAttention, FFN
from eight_mile.tf.embeddings import LookupTableEmbeddings, LearnedPositionalLookupTableEmbeddingsWithBias, LearnedPositionalLookupTableEmbeddings

import tensorflow as tf


def from_weight_array(tf_layer: tf.keras.layers.Layer, d: Dict, name: str):
    """Read in {`LayerNorm`, `Linear`, `layers.Dense`} from `weights` and `bias` fields

    :param tf_layer: A layer to get weights for
    :param d: A Dict containing the arrays by key
    :param name: The name of this layer
    :return: None
    """
    weights = d[f"{name}/weights"]
    bias = d[f"{name}/bias"]
    tf_layer.set_weights([weights.T, bias])


def from_ffn_array(tf_ffn: FFN, d: Dict, name: str):
    """Restore an `FFN` layer's weights from a set of arrays

    :param tf_ffn: An `FFN` layer for Transformers
    :param d: A Dict containing the arrays by key
    :param name: The name of the layer
    :return: None
    """
    from_weight_array(tf_ffn.expansion, d, f"{name}/expansion")
    from_weight_array(tf_ffn.squeeze, d, f"{name}/squeeze")


def from_attn_array(tf_attn: tf.keras.layers.Layer, d: Dict, name: str):
    """Restore a self-attention module from a set of keys

    :param tf_attn: A self-attention module for Transformers, could be MultiHeadedAttention or
    MultiHeadedRelativeAttention
    :param d: A Dict of arrays by key
    :param name: The name of the layer
    """
    from_weight_array(tf_attn.w_Q, d, f"{name}/w_Q")
    from_weight_array(tf_attn.w_K, d, f"{name}/w_K")
    from_weight_array(tf_attn.w_V, d, f"{name}/w_V")
    from_weight_array(tf_attn.w_O, d, f"{name}/w_O")

    if hasattr(tf_attn, 'rpr_key'):
        tf_attn.rpr_key.set_weights([d[f"{name}/rpr_key"]])
        tf_attn.rpr_value.set_weights([d[f"{name}/rpr_value"]])


def from_encoder_array(tf_encoder: TransformerEncoder, d: Dict, name: str):
    """Restore a `TransformerEncoder` layer from a set of numpy arrays

    :param tf_encoder: A `TransformerEncoder` layer
    :param d: A Dict of arrays by key
    :param name: The layer name
    :return: None
    """
    from_weight_array(tf_encoder.ln1, d, f"{name}/ln1")
    from_weight_array(tf_encoder.ln2, d, f"{name}/ln2")
    from_attn_array(tf_encoder.self_attn, d, f"{name}/attn")
    from_ffn_array(tf_encoder.ffn, d, f"{name}/ffn")


def from_embed_array(tf_embed: tf.keras.layers.Layer, d: Dict, name: str, bias=None):
    """Restore a simple lookup table embedding from a `weights` array

    :param tf_embed: An embedding module
    :param d: A Dict containing a `weights` array to restore
    :param name: name of the layer
    :return: None
    """
    weights = [d[f"{name}/weights"]]
    if hasattr(tf_embed, 'pos'):
        pos_weights = d[f"{name}/pos_weights"]
        weights = [pos_weights] + weights
        if hasattr(tf_embed, 'bias') and bias is not None:
            weights = weights + [bias.reshape(1, -1)]

    tf_embed.set_weights(weights)


def from_encoder_stack_array(tf_encoder_stack: TransformerEncoderStack, d: Dict, name: str = "TransformerEncoderStack"):
    """Restore weights from a `TransformerEncoderStack`

    :param tf_encoder_stack: A transformer encoder stack
    :param d: A Dict containing sets of arrays
    :param name: A name for this primitive
    :return: None
    """
    if isinstance(tf_encoder_stack.ln, tf.keras.layers.LayerNormalization):
        from_weight_array(tf_encoder_stack.ln, d, f"{name}/ln")
    for i, enc_tf in enumerate(tf_encoder_stack.encoders):
        from_encoder_array(enc_tf, d, f"{name}/{i}")


def from_tlm_array(tf_tlm: tf.keras.layers.Layer, d: Dict, embeddings_keys: List[str] = None, name: str = "TLM"):
    """Restore a TLM-like model (possibly a `Model` for fine-tuning)

    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
    uninitialized

    :param tf_tlm: A TLM-like model
    :param d: A Dict of weights to restore for each layer
    :param embeddings_keys: Name of the embeddings to restore, defaults to `None` in which case restore all embeddings
    :param name: A name for this primitive
    :return:
    """
    transformer = tf_tlm.transformer if hasattr(tf_tlm, 'transformer') else tf_tlm.generator
    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")

    keys_to_restore = embeddings_keys if embeddings_keys else list(tf_tlm.embeddings.keys())
    for embeddings_key in keys_to_restore:
        # If we get this class in we have to monkey patch the embeddings so we can add the bias
        if isinstance(tf_tlm.embeddings[embeddings_key], LearnedPositionalLookupTableEmbeddingsWithBias):
            bias = d[f"{name}/Embeddings/tt/weights"][0]
            from_embed_array(tf_tlm.embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}", bias)
        else:
            from_embed_array(tf_tlm.embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
    if hasattr(tf_tlm.embeddings.reduction, 'ln'):
        from_weight_array(tf_tlm.embeddings.reduction.ln, d, f"{name}/Embeddings/reduction/ln")


def load_tlm_npz(tf_tlm: tf.keras.layers.Layer, npz: str, embeddings_key: str = 'x', name: str = "TLM"):
    """Restore a TLM-like model (possibly a `Model` for fine-tuning

    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
    uninitialized

    :param tf_tlm: A TLM-like model
    :param npz: A file to restore the weights from
    :param embeddings_key: The name of the embeddings to restore, defaults to `x`
    :param name: A name for this primitive
    :return:
    """
    d = np.load(npz)
    from_tlm_array(tf_tlm, d, embeddings_key, name)



import math
import numpy as np
from eight_mile.utils import exporter, listify
from eight_mile.utils import optional_params, register


__all__ = []
export = exporter(__all__)

MEAD_LAYERS_LR_SCHEDULERS = {}
export = exporter(__all__)


@export
@optional_params
def register_lr_scheduler(cls, name=None):
    return register(cls, MEAD_LAYERS_LR_SCHEDULERS, name, "lr_scheduler")


@export
def create_lr_scheduler(**kwargs):
    """Create a learning rate scheduler.

    :Keyword Arguments:
      * *lr_scheduler_type* `str` or `list` The name of the learning rate scheduler
          if list then the first scheduler should be a warmup scheduler.
    """
    sched_type = kwargs.get("lr_scheduler_type")
    if sched_type is None:
        return None
    sched_type = listify(sched_type)
    if len(sched_type) == 2:
        warm = MEAD_LAYERS_LR_SCHEDULERS.get(sched_type[0])(**kwargs)
        assert isinstance(warm, WarmupLearningRateScheduler), "First LR Scheduler must be a warmup scheduler."
        rest = MEAD_LAYERS_LR_SCHEDULERS.get(sched_type[1])(**kwargs)
        return MEAD_LAYERS_LR_SCHEDULERS.get("composite")(warm=warm, rest=rest, **kwargs)
    Constructor = MEAD_LAYERS_LR_SCHEDULERS.get(sched_type[0])
    return Constructor(**kwargs)


@export
class LearningRateScheduler(object):
    def __init__(self, **kwargs):
        self.lr = kwargs.get("lr", kwargs.get("eta", 1.0))

    @staticmethod
    def _identity(x):
        return x


@export
class WarmupLearningRateScheduler(LearningRateScheduler):
    def __init__(self, warmup_steps=16000, **kwargs):
        super().__init__(**kwargs)
        self._warmup_steps = warmup_steps

    @property
    def warmup_steps(self):
        return self._warmup_steps


@export
class ConstantScheduler(LearningRateScheduler):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, global_step):
        return self.lr


@export
class WarmupLinearScheduler(WarmupLearningRateScheduler):
    def __call__(self, global_step):
        lr_factor = min(1.0, global_step / float(self.warmup_steps))
        return self.lr * lr_factor


@export
class CyclicLRScheduler(LearningRateScheduler):
    def __init__(self, max_lr=1e-2, decay_steps=1000, **kwargs):
        super().__init__(**kwargs)
        self.max_lr = max_lr
        self.decay_steps = decay_steps

    def __call__(self, global_step):
        cycle = np.floor(1.0 + global_step / (2.0 * self.decay_steps))
        x = np.abs(global_step / self.decay_steps - 2.0 * cycle + 1.0)
        new_lr = self.lr + (self.max_lr - self.lr) * np.maximum(0.0, 1.0 - x)
        return new_lr


@export
class PiecewiseDecayScheduler(LearningRateScheduler):
    def __init__(self, boundaries, values, **kwargs):
        super().__init__(**kwargs)
        self.boundaries = boundaries
        self.values = values

    def __call__(self, global_step):
        pos = np.searchsorted(self.boundaries, global_step)
        return self.values[pos]


@export
class ZarembaDecayScheduler(PiecewiseDecayScheduler):
    def __init__(self, boundaries=None, decay_rate=None, **kwargs):
        lr = kwargs.get("lr", kwargs.get("eta", 1.0))

        if boundaries is None or decay_rate is None:
            boundaries = []
            values = [lr]
        else:
            values = [lr / (decay_rate ** i) for i in range(len(boundaries) + 1)]
        super().__init__(boundaries, values, **kwargs)


@export
class CosineDecayScheduler(LearningRateScheduler):
    def __init__(self, decay_steps=1000, alpha=0.0, **kwargs):
        super().__init__(**kwargs)
        self.decay_steps = decay_steps
        self.alpha = alpha

    def __call__(self, global_step):
        global_step = min(global_step, self.decay_steps)
        cosine_decay = 0.5 * (1 + np.cos(np.pi * global_step / self.decay_steps))
        decayed = (1 - self.alpha) * cosine_decay + self.alpha
        return self.lr * decayed


@export
class InverseTimeDecayScheduler(LearningRateScheduler):
    def __init__(self, decay_steps=16000, decay_rate=0.05, staircase=False, **kwargs):
        super().__init__(**kwargs)
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate
        self.wrap_fn = math.floor if staircase else LearningRateScheduler._identity

    def __call__(self, global_step):
        t = self.wrap_fn(global_step / self.decay_steps)
        return self.lr / (1.0 + self.decay_rate * t)


@export
class ExponentialDecayScheduler(LearningRateScheduler):
    def __init__(self, decay_steps=16000, decay_rate=0.5, staircase=False, **kwargs):
        super().__init__(**kwargs)
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate
        self.wrap_fn = math.floor if staircase else LearningRateScheduler._identity

    def __call__(self, global_step):
        t = self.wrap_fn(global_step / float(self.decay_steps))
        return self.lr * self.decay_rate ** t


@export
class CompositeLRScheduler(LearningRateScheduler):
    def __init__(self, warm=None, rest=None, **kwargs):
        super().__init__(**kwargs)
        self.warm = warm
        self.rest = rest

    def __call__(self, global_step):
        if global_step < self.warm.warmup_steps:
            return self.warm(global_step)
        return self.rest(global_step - self.warm.warmup_steps)
